from sqlalchemy import case, cast, Float, func

import vigorish.database as db

avg_numerator = func.sum(db.BatStats.hits)
avg_denominator = func.sum(db.BatStats.at_bats)
avg_ = avg_numerator / cast(avg_denominator, Float)
avg = case([(avg_denominator > 0, avg_)], else_=0.0).label("avg")

obp_numerator = func.sum(db.BatStats.hits) + func.sum(db.BatStats.bases_on_balls) + func.sum(db.BatStats.hit_by_pitch)
obp_denominator = (
    func.sum(db.BatStats.at_bats)
    + func.sum(db.BatStats.bases_on_balls)
    + func.sum(db.BatStats.hit_by_pitch)
    + func.sum(db.BatStats.sac_fly)
)
obp_ = obp_numerator / cast(obp_denominator, Float)
obp = case([(obp_denominator > 0, obp_)], else_=0.0).label("obp")

extra_base_hits = func.sum(db.BatStats.doubles) + func.sum(db.BatStats.triples) + func.sum(db.BatStats.homeruns)
singles = func.sum(db.BatStats.hits) - extra_base_hits
slg_numerator = (
    singles
    + (func.sum(db.BatStats.doubles) * 2)
    + (func.sum(db.BatStats.triples) * 3)
    + (func.sum(db.BatStats.homeruns) * 4)
)
slg_denominator = func.sum(db.BatStats.at_bats)
slg_ = slg_numerator / cast(slg_denominator, Float)
slg = case([(slg_denominator > 0, slg_)], else_=0.0).label("slg")

ops = (obp + slg).label("ops")
iso = (slg - avg).label("iso")

bb_rate_numerator = func.sum(db.BatStats.bases_on_balls)
bb_rate_denominator = func.sum(db.BatStats.plate_appearances)
bb_rate_ = bb_rate_numerator / cast(bb_rate_denominator, Float)
bb_rate = case([(bb_rate_denominator > 0, bb_rate_)], else_=0.0).label("bb_rate")

k_rate_numerator = func.sum(db.BatStats.strikeouts)
k_rate_denominator = func.sum(db.BatStats.plate_appearances)
k_rate_ = k_rate_numerator / cast(k_rate_denominator, Float)
k_rate = case([(k_rate_denominator > 0, k_rate_)], else_=0.0).label("k_rate")

contact_rate_numerator = func.sum(db.BatStats.at_bats) - func.sum(db.BatStats.strikeouts)
contact_rate_denominator = func.sum(db.BatStats.at_bats)
contact_rate_ = contact_rate_numerator / cast(contact_rate_denominator, Float)
contact_rate = case([(contact_rate_denominator > 0, contact_rate_)], else_=0.0).label("contact_rate")



from dataclasses import dataclass


@dataclass
class BBRefGameInfo:
    """Individual game info scraped from brooksbaseball.com."""

    url: str = ""
    bbref_game_id: str = ""

    def as_dict(self):
        """Convert game info to a dictionary."""
        return {
            "__bbref_game_info__": True,
            "url": self.url,
            "bbref_game_id": self.bbref_game_id,
        }



"""Functions for reading and writing files."""
import json
import os
from pathlib import Path

import boto3
import botocore

from vigorish.data.json_decoder import (
    decode_bbref_boxscore,
    decode_bbref_boxscore_patch_list,
    decode_bbref_games_for_date,
    decode_bbref_games_for_date_patch_list,
    decode_brooks_games_for_date,
    decode_brooks_games_for_date_patch_list,
    decode_brooks_pitch_logs_for_game,
    decode_brooks_pitchfx_log,
    decode_brooks_pitchfx_patch_list,
)
from vigorish.enums import DataSet, LocalFileTask, S3FileTask, VigFile
from vigorish.util.dt_format_strings import DATE_ONLY, DATE_ONLY_TABLE_ID
from vigorish.util.exceptions import S3BucketException
from vigorish.util.numeric_helpers import ONE_KB
from vigorish.util.result import Result


class FileHelper:
    def __init__(self, config):
        self.config = config
        self.bucket_name = self.config.get_current_setting("S3_BUCKET", DataSet.ALL)
        self.s3_resource = None
        try:  # pragma: no cover
            if os.environ.get("ENV") != "TEST":
                self.s3_resource = boto3.resource("s3")
        except (boto3.exceptions.ResourceNotExistsError, ValueError):  # pragma: no cover
            self.s3_resource = None

    @property
    def local_folderpath_dict(self):
        html_local_folder = self.config.all_settings.get("HTML_LOCAL_FOLDER_PATH")
        json_local_folder = self.config.all_settings.get("JSON_LOCAL_FOLDER_PATH")
        combined_local_folder = self.config.all_settings.get("COMBINED_DATA_LOCAL_FOLDER_PATH")
        return {
            VigFile.SCRAPED_HTML: {ds: html_local_folder.current_setting(data_set=ds) for ds in DataSet},
            VigFile.PARSED_JSON: {ds: json_local_folder.current_setting(data_set=ds) for ds in DataSet},
            VigFile.PATCH_LIST: {ds: json_local_folder.current_setting(data_set=ds) for ds in DataSet},
            VigFile.COMBINED_GAME_DATA: {DataSet.ALL: combined_local_folder.current_setting(data_set=DataSet.ALL)},
        }

    @property
    def s3_folderpath_dict(self):  # pragma: no cover
        html_s3_folder = self.config.all_settings.get("HTML_S3_FOLDER_PATH")
        json_s3_folder = self.config.all_settings.get("JSON_S3_FOLDER_PATH")
        combined_s3_folder = self.config.all_settings.get("COMBINED_DATA_S3_FOLDER_PATH")
        return {
            VigFile.SCRAPED_HTML: {ds: html_s3_folder.current_setting(data_set=ds) for ds in DataSet},
            VigFile.PARSED_JSON: {ds: json_s3_folder.current_setting(data_set=ds) for ds in DataSet},
            VigFile.PATCH_LIST: {ds: json_s3_folder.current_setting(data_set=ds) for ds in DataSet},
            VigFile.COMBINED_GAME_DATA: {DataSet.ALL: combined_s3_folder.current_setting(data_set=DataSet.ALL)},
        }

    @property
    def filename_dict(self):
        html_filename_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.get_file_name_html_brooks_games_for_date,
            DataSet.BROOKS_PITCH_LOGS: self.get_file_name_html_brooks_pitch_log,
            DataSet.BROOKS_PITCHFX: self.get_file_name_html_brooks_pitchfx,
            DataSet.BBREF_GAMES_FOR_DATE: self.get_file_name_html_bbref_games_for_date,
            DataSet.BBREF_BOXSCORES: self.get_file_name_html_bbref_boxscore,
        }
        json_filename_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.get_file_name_json_brooks_games_for_date,
            DataSet.BROOKS_PITCH_LOGS: self.get_file_name_json_brooks_pitch_log_for_game,
            DataSet.BROOKS_PITCHFX: self.get_file_name_json_brooks_pitchfx,
            DataSet.BBREF_GAMES_FOR_DATE: self.get_file_name_json_bbref_games_for_date,
            DataSet.BBREF_BOXSCORES: self.get_file_name_json_bbref_boxscore,
        }
        patch_list_filename_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.get_file_name_brooks_games_for_date_patch_list,
            DataSet.BROOKS_PITCHFX: self.get_file_name_brooks_pitchfx_patch_list,
            DataSet.BBREF_GAMES_FOR_DATE: self.get_file_name_bbref_games_for_date_patch_list,
            DataSet.BBREF_BOXSCORES: self.get_file_name_bbref_boxscore_patch_list,
        }
        return {
            VigFile.SCRAPED_HTML: html_filename_dict,
            VigFile.PARSED_JSON: json_filename_dict,
            VigFile.PATCH_LIST: patch_list_filename_dict,
            VigFile.COMBINED_GAME_DATA: {DataSet.ALL: self.get_file_name_combined_game_data},
        }

    @property
    def json_decoder_dict(self):
        scraped_data_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: decode_brooks_games_for_date,
            DataSet.BROOKS_PITCH_LOGS: decode_brooks_pitch_logs_for_game,
            DataSet.BROOKS_PITCHFX: decode_brooks_pitchfx_log,
            DataSet.BBREF_GAMES_FOR_DATE: decode_bbref_games_for_date,
            DataSet.BBREF_BOXSCORES: decode_bbref_boxscore,
        }
        patch_file_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: decode_brooks_games_for_date_patch_list,
            DataSet.BROOKS_PITCHFX: decode_brooks_pitchfx_patch_list,
            DataSet.BBREF_GAMES_FOR_DATE: decode_bbref_games_for_date_patch_list,
            DataSet.BBREF_BOXSCORES: decode_bbref_boxscore_patch_list,
        }
        return {
            VigFile.PARSED_JSON: scraped_data_dict,
            VigFile.PATCH_LIST: patch_file_dict,
        }

    @property
    def file_storage_dict(self):
        html_storage = self.config.all_settings.get("HTML_STORAGE")
        json_storage = self.config.all_settings.get("JSON_STORAGE")
        combined_storage = self.config.all_settings.get("COMBINED_DATA_STORAGE")
        return {
            VigFile.SCRAPED_HTML: {ds: html_storage.current_setting(data_set=ds) for ds in DataSet},
            VigFile.PARSED_JSON: {ds: json_storage.current_setting(data_set=ds) for ds in DataSet},
            VigFile.PATCH_LIST: {ds: json_storage.current_setting(data_set=ds) for ds in DataSet},
            VigFile.COMBINED_GAME_DATA: {DataSet.ALL: combined_storage.current_setting(data_set=DataSet.ALL)},
        }

    def create_all_folderpaths(self, year):
        try:
            for folderpath_dict in self.local_folderpath_dict.values():
                for folderpath in folderpath_dict.values():
                    Path(folderpath.resolve(year=year)).mkdir(parents=True, exist_ok=True)
            return Result.Ok()
        except Exception as e:
            return Result.Fail(f"Error: {repr(e)}")

    def check_s3_bucket(self):  # pragma: no cover
        if not self.s3_resource:
            raise S3BucketException()
        try:
            s3_bucket = self.s3_resource.Bucket(self.bucket_name) if self.bucket_name else None
            return Result.Ok(s3_bucket)
        except botocore.exceptions.ClientError as ex:
            raise S3BucketException(f'{repr(ex)} (Error Code: {ex.response["Error"]["Code"]})')

    def check_file_stored_local(self, file_type, data_set):
        storage_setting = self.file_storage_dict[file_type][data_set]
        return "LOCAL_FOLDER" in storage_setting.name or "BOTH" in storage_setting.name

    def check_file_stored_s3(self, file_type, data_set):  # pragma: no cover
        storage_setting = self.file_storage_dict[file_type][data_set]
        return "S3_BUCKET" in storage_setting.name or "BOTH" in storage_setting.name

    def get_s3_bucket(self):  # pragma: no cover
        result = self.check_s3_bucket()
        return None if result.failure else result.value

    def get_all_object_keys_in_s3_bucket(self):  # pragma: no cover
        s3_bucket = self.get_s3_bucket()
        return list(s3_bucket.objects.all()) if s3_bucket else []

    def perform_local_file_task(
        self,
        task,
        data_set,
        file_type,
        game_date,
        scraped_data=None,
        bbref_game_id=None,
        bb_game_id=None,
        pitch_app_id=None,
        delete_file=False,
    ):
        filepath = self.get_local_filepath(
            data_set=data_set,
            file_type=file_type,
            game_date=game_date,
            bbref_game_id=bbref_game_id,
            bb_game_id=bb_game_id,
            pitch_app_id=pitch_app_id,
        )
        if task == LocalFileTask.READ_FILE:
            return self.read_local_file(Path(filepath))
        if task == LocalFileTask.WRITE_FILE:
            return self.write_to_file(file_type, scraped_data, Path(filepath))
        if task == LocalFileTask.DELETE_FILE:
            return self.delete_file(Path(filepath))
        if task == LocalFileTask.DECODE_JSON:
            return self.decode_json(file_type, data_set, Path(filepath))

    def get_local_filepath(
        self,
        data_set,
        file_type,
        game_date,
        bbref_game_id=None,
        bb_game_id=None,
        pitch_app_id=None,
    ):
        folderpath = self.get_local_folderpath(file_type, data_set, game_date)
        filename = self.get_file_name(
            data_set,
            file_type,
            game_date,
            bbref_game_id,
            bb_game_id,
            pitch_app_id,
        )
        return str(Path(folderpath).joinpath(filename))

    def get_local_folderpath(self, file_type, data_set, game_date=None, year=None):
        if not game_date and not year:
            error = "You must provide either the game_date or year argument to construct a folderpath"
            raise ValueError(error)
        year = year or game_date.year
        return self.local_folderpath_dict[file_type][data_set].resolve(year=year)

    def get_file_name(
        self,
        data_set,
        file_type,
        game_date,
        bbref_game_id=None,
        bb_game_id=None,
        pitch_app_id=None,
    ):
        if pitch_app_id:
            url_id = pitch_app_id
        elif bb_game_id:
            url_id = bb_game_id
        elif bbref_game_id:
            url_id = bbref_game_id
        elif game_date:
            url_id = game_date
        else:
            raise ValueError("Unable to construct file name.")
        return self.filename_dict[file_type][data_set](url_id) if data_set in self.filename_dict[file_type] else None

    def perform_s3_task(
        self,
        task,
        data_set,
        file_type,
        game_date,
        scraped_data=None,
        bbref_game_id=None,
        bb_game_id=None,
        pitch_app_id=None,
    ):  # pragma: no cover
        s3_key = self.get_s3_object_key(
            data_set=data_set,
            file_type=file_type,
            game_date=game_date,
            bbref_game_id=bbref_game_id,
            bb_game_id=bb_game_id,
            pitch_app_id=pitch_app_id,
        )
        filepath = self.get_local_filepath(
            data_set=data_set,
            file_type=file_type,
            game_date=game_date,
            bbref_game_id=bbref_game_id,
            bb_game_id=bb_game_id,
            pitch_app_id=pitch_app_id,
        )
        if task == S3FileTask.UPLOAD:
            return self.upload_to_s3(file_type, data_set, scraped_data, s3_key, Path(filepath))
        if task == S3FileTask.DOWNLOAD:
            return self.download_from_s3(file_type, s3_key, Path(filepath))
        if task == S3FileTask.DELETE:
            return self.delete_from_s3(s3_key)

    def get_s3_object_key(
        self,
        data_set,
        file_type,
        game_date,
        bbref_game_id=None,
        bb_game_id=None,
        pitch_app_id=None,
    ):  # pragma: no cover
        folderpath = self.get_s3_folderpath(file_type, data_set, game_date)
        filename = self.get_file_name(
            data_set=data_set,
            file_type=file_type,
            game_date=game_date,
            bbref_game_id=bbref_game_id,
            bb_game_id=bb_game_id,
            pitch_app_id=pitch_app_id,
        )
        return f"{folderpath}/{filename}"

    def get_s3_folderpath(self, file_type, data_set, game_date=None, year=None):  # pragma: no cover
        if not game_date and not year:
            error = "You must provide either the game_date or year argument to construct a folderpath"
            raise ValueError(error)
        year = year or game_date.year
        return self.s3_folderpath_dict[file_type][data_set].resolve(year=year)

    def get_file_name_html_brooks_games_for_date(self, game_date):
        return f"{game_date.strftime(DATE_ONLY_TABLE_ID)}.html"

    def get_file_name_json_brooks_games_for_date(self, game_date):
        return f"brooks_games_for_date_{game_date.strftime(DATE_ONLY)}.json"

    def get_file_name_brooks_games_for_date_patch_list(self, game_date):
        return f"brooks_games_for_date_{game_date.strftime(DATE_ONLY)}_PATCH_LIST.json"

    def get_file_name_html_brooks_pitch_log(self, pitch_app_id):
        return f"{pitch_app_id}.html"

    def get_file_name_json_brooks_pitch_log_for_game(self, bb_game_id):
        return f"{bb_game_id}.json"

    def get_file_name_html_brooks_pitchfx(self, pitch_app_id):
        return f"{pitch_app_id}.html"

    def get_file_name_json_brooks_pitchfx(self, pitch_app_id):
        return f"{pitch_app_id}.json"

    def get_file_name_brooks_pitchfx_patch_list(self, bbref_game_id):
        return f"{bbref_game_id}_PATCH_LIST.json"

    def get_file_name_html_bbref_games_for_date(self, game_date):
        return f"{game_date.strftime(DATE_ONLY_TABLE_ID)}.html"

    def get_file_name_json_bbref_games_for_date(self, game_date):
        return f"bbref_games_for_date_{game_date.strftime(DATE_ONLY)}.json"

    def get_file_name_bbref_games_for_date_patch_list(self, game_date):
        return f"bbref_games_for_date_{game_date.strftime(DATE_ONLY)}_PATCH_LIST.json"

    def get_file_name_html_bbref_boxscore(self, bbref_game_id):
        return f"{bbref_game_id}.html"

    def get_file_name_json_bbref_boxscore(self, bbref_game_id):
        return f"{bbref_game_id}.json"

    def get_file_name_bbref_boxscore_patch_list(self, bbref_game_id):
        return f"{bbref_game_id}_PATCH_LIST.json"

    def get_file_name_combined_game_data(self, bbref_game_id):
        return f"{bbref_game_id}_COMBINED_DATA.json"

    def read_local_file(self, filepath):
        return Result.Ok(filepath) if filepath.exists() else Result.Fail(f"File not found: {filepath.resolve()}.")

    def write_to_file(self, file_type, data, filepath):
        """Write object in json format to file."""
        if file_type in [VigFile.PARSED_JSON, VigFile.PATCH_LIST]:
            data = data.as_json()
        if file_type == VigFile.COMBINED_GAME_DATA:
            data = json.dumps(data, indent=2, sort_keys=False)
        try:
            filepath.write_text(data)
            return Result.Ok(filepath)
        except Exception as e:
            error = f"Error: {repr(e)}"
            return Result.Fail(error)

    def delete_file(self, filepath):
        if filepath.exists() and filepath.is_file():
            filepath.unlink()
        return Result.Ok()

    def decode_json(self, file_type, data_set, filepath):
        delete_file = not self.check_file_stored_local(VigFile.PARSED_JSON, data_set)
        try:
            contents = filepath.read_text()
            if delete_file and os.environ.get("ENV") != "TEST":
                filepath.unlink()
            return self.json_decoder_dict[file_type][data_set](json.loads(contents))
        except Exception as e:
            error = f"Error: {repr(e)}"
            return Result.Fail(error)

    def upload_to_s3(self, file_type, data_set, scraped_data, s3_key, filepath):  # pragma: no cover
        delete_file = not self.check_file_stored_local(file_type, data_set)
        if file_type == VigFile.PARSED_JSON:
            result = self.write_to_file(file_type, scraped_data, filepath)
            if result.failure:
                return result
        try:
            self.get_s3_bucket().upload_file(str(filepath), s3_key)
            if delete_file:
                filepath.unlink()
            return Result.Ok() if delete_file else Result.Ok(filepath)
        except botocore.exceptions.ClientError as ex:
            error_code = ex.response["Error"]["Code"]
            return Result.Fail(f"{repr(ex)} (Error Code: {error_code})")

    def download_from_s3(self, file_type, s3_key, filepath):  # pragma: no cover
        try:
            self.get_s3_bucket().download_file(s3_key, str(filepath))
            if file_type == VigFile.SCRAPED_HTML and filepath.stat().st_size < ONE_KB:
                self.s3_resource.Object(self.bucket_name, s3_key).delete()
                filepath.unlink()
                error = f"Size of file downloaded from S3 is less than 1KB ({s3_key})"
                return Result.Fail(error)
            return Result.Ok(filepath)
        except botocore.exceptions.ClientError as ex:
            error_code = ex.response["Error"]["Code"]
            if error_code == "404":
                error = f'The object "{s3_key}" does not exist.'
            else:
                error = f"{repr(ex)} (Error Code: {error_code})"
            return Result.Fail(error)

    def delete_from_s3(self, s3_key):  # pragma: no cover
        try:
            self.s3_resource.Object(self.bucket_name, s3_key).delete()
            return Result.Ok()
        except botocore.exceptions.ClientError as ex:
            error_code = ex.response["Error"]["Code"]
            return Result.Fail(f"{repr(ex)} (Error Code: {error_code})")

    def rename_s3_object(self, old_key, new_key):  # pragma: no cover
        try:
            self.s3_resource.Object(self.bucket_name, new_key).copy_from(CopySource=f"{self.bucket_name}/{old_key}")
            self.s3_resource.Object(self.bucket_name, old_key).delete()
            return Result.Ok()
        except botocore.exceptions.ClientError as ex:
            error_code = ex.response["Error"]["Code"]
            return Result.Fail(f"{repr(ex)} (Error Code: {error_code})")



import json
from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Mapping

from dateutil import tz

from vigorish.scrape.brooks_pitchfx.models.pitchfx import BrooksPitchFxData
from vigorish.util.list_helpers import as_dict_list


@dataclass
class BrooksPitchFxLog:
    pitch_count_by_inning: Mapping
    total_pitch_count: str = "0"
    pitcher_name: str = ""
    pitcher_id_mlb: str = "0"
    pitch_app_id: str = ""
    pitcher_team_id_bb: str = ""
    opponent_team_id_bb: str = ""
    bb_game_id: str = ""
    bbref_game_id: str = ""
    game_date_year: str = ""
    game_date_month: str = ""
    game_date_day: str = ""
    game_time_hour: str = ""
    game_time_minute: str = ""
    time_zone_name: str = ""
    pitchfx_url: str = ""
    pitchfx_log: List[BrooksPitchFxData] = field(default_factory=list)

    @property
    def game_date(self):
        return datetime(self.game_date_year, self.game_date_month, self.game_date_day).date()

    @property
    def game_start_time(self):
        if not self.game_time_hour:
            return None
        game_start_time = datetime(
            year=self.game_date.year,
            month=self.game_date.month,
            day=self.game_date.day,
            hour=self.game_time_hour,
            minute=self.game_time_minute,
        )
        return game_start_time.replace(tzinfo=tz.gettz(self.time_zone_name))

    def as_dict(self):
        return {
            "__brooks_pitchfx_log__": True,
            "pitchfx_log": as_dict_list(self.pitchfx_log),
            "pitch_count_by_inning": self.pitch_count_by_inning,
            "pitcher_name": self.pitcher_name,
            "pitcher_id_mlb": int(self.pitcher_id_mlb),
            "pitch_app_id": self.pitch_app_id,
            "total_pitch_count": int(self.total_pitch_count),
            "pitcher_team_id_bb": self.pitcher_team_id_bb,
            "opponent_team_id_bb": self.opponent_team_id_bb,
            "bb_game_id": self.bb_game_id,
            "bbref_game_id": self.bbref_game_id,
            "game_date_year": int(self.game_date_year),
            "game_date_month": int(self.game_date_month),
            "game_date_day": int(self.game_date_day),
            "game_time_hour": int(self.game_time_hour),
            "game_time_minute": int(self.game_time_minute),
            "time_zone_name": self.time_zone_name,
            "pitchfx_url": self.pitchfx_url,
        }

    def as_json(self):
        """Convert pitchfx log to JSON."""
        return json.dumps(self.as_dict(), indent=2)



from collections import defaultdict, OrderedDict
from copy import deepcopy
from dataclasses import dataclass, field
from typing import List

from dacite import from_dict

import vigorish.database as db
from vigorish.enums import DataSet
from vigorish.patch.base import Patch, PatchList
from vigorish.scrape.brooks_pitchfx.models.pitchfx_log import BrooksPitchFxLog
from vigorish.util.list_helpers import group_and_sort_list
from vigorish.util.result import Result
from vigorish.util.string_helpers import get_brooks_team_id


@dataclass
class BrooksPitchFxPatchList(PatchList):
    original_pfx_logs: List[BrooksPitchFxLog] = field(repr=False, init=False)

    def __post_init__(self):
        self.data_set = DataSet.BROOKS_PITCHFX
        self.patch_list_id = "__brooks_pitchfx_patch_list__"

    def apply(self, data, db_session, boxscore):
        data = self.pre_process_data(data)
        for patch in self.patch_list:
            result = patch.apply(data)
            if result.failure:
                return result
            data = result.value
        result = self.post_process_data(data, db_session, boxscore)
        if result.failure:
            return result
        data = result.value
        return Result.Ok(data)

    def pre_process_data(self, data):
        self.original_pfx_logs = {pfx_log.pitch_app_id: pfx_log for pfx_log in deepcopy(data)}
        return [pfx for pitchfx_log in data for pfx in pitchfx_log.pitchfx_log]

    def post_process_data(self, data, db_session, boxscore=None):
        pfx_grouped = group_and_sort_list(data, "pitch_app_id", "park_sv_id")
        for pa_id, pfx_log in pfx_grouped.items():
            og_pfx_log = self.original_pfx_logs.get(pa_id, None)
            if not og_pfx_log:
                new_pfx_log = self.create_new_pitchfx_log(pa_id, pfx_log, db_session, boxscore)
                self.original_pfx_logs[pa_id] = new_pfx_log
                continue
            og_pfx_log.pitchfx_log = pfx_log
            og_pfx_log.pitch_count_by_inning = self.get_pitch_count_by_inning(pfx_log)
            og_pfx_log.total_pitch_count = len(pfx_log)
        patched_pfx_logs = list(self.original_pfx_logs.values())
        return Result.Ok(patched_pfx_logs)

    def get_pitch_count_by_inning(self, pitchfx_log):
        pitch_count_unordered = defaultdict(int)
        for pfx in pitchfx_log:
            pitch_count_unordered[pfx.inning] += 1
        pitch_count_ordered = OrderedDict()
        for k in sorted(pitch_count_unordered.keys()):
            pitch_count_ordered[k] = pitch_count_unordered[k]
        return pitch_count_ordered

    def create_new_pitchfx_log(self, pitch_app_id, pitchfx_log, db_session, boxscore):
        pitcher_id_mlb = pitch_app_id.split("_")[1]
        player = db.PlayerId.find_by_mlb_id(db_session, pitcher_id_mlb)
        away_team_id = boxscore.away_team_data.team_id_br
        home_team_id = boxscore.home_team_data.team_id_br
        player_team_id = boxscore.player_team_dict[player.bbref_id]
        opponent_team_id = away_team_id if player_team_id == home_team_id else home_team_id
        game_status = db.GameScrapeStatus.find_by_bbref_game_id(db_session, self.url_id)
        pfx_log_dict = {
            "pitchfx_log": pitchfx_log,
            "pitch_count_by_inning": self.get_pitch_count_by_inning(pitchfx_log),
            "total_pitch_count": str(len(pitchfx_log)),
            "pitcher_name": player.mlb_name,
            "pitcher_id_mlb": pitcher_id_mlb,
            "pitch_app_id": pitch_app_id,
            "pitcher_team_id_bb": get_brooks_team_id(player_team_id),
            "opponent_team_id_bb": get_brooks_team_id(opponent_team_id),
            "bb_game_id": game_status.bb_game_id,
            "bbref_game_id": self.url_id,
            "game_date_year": str(game_status.game_date.year),
            "game_date_month": str(game_status.game_date.month),
            "game_date_day": str(game_status.game_date.day),
            "game_date_hour": str(game_status.game_time_hour),
            "game_date_minute": str(game_status.game_time_minute),
            "time_zone_name": game_status.game_time_zone,
            "pitchfx_url": "",
        }
        pfx_log = from_dict(data_class=BrooksPitchFxLog, data=pfx_log_dict)
        pfx_log.pitcher_id_mlb = int(pitcher_id_mlb)
        return pfx_log


@dataclass
class PatchBrooksPitchFxBatterId(Patch):
    bbref_game_id: str
    park_sv_id: str = field(repr=False)
    pitch_app_id: str = field(repr=False)
    current_at_bat_id: str = field(repr=False)
    current_batter_id: int
    current_batter_name: str = field(repr=False)
    new_at_bat_id: str = field(repr=False)
    new_batter_id: int
    new_batter_name: str = field(repr=False)

    def __post_init__(self):
        self.data_set = DataSet.BROOKS_PITCHFX
        self.patch_id = "__patch_brooks_pitchfx_batter_id__"

    def apply(self, data):
        matches = [pfx for pfx in data if pfx.park_sv_id == self.park_sv_id and pfx.pitch_app_id == self.pitch_app_id]
        if not matches:
            error = (
                "Unable to locate the PitchFX reading identified in this patch: "
                f"pitch_app_id: {self.pitch_app_id}, park_sv_id: {self.park_sv_id}"
            )
            return Result.Fail(error)
        if len(matches) > 1:
            error = (
                "More than one PitchFX raeding was found that matches this patch: "
                f"pitch_app_id: {self.pitch_app_id}, park_sv_id: {self.park_sv_id}"
            )
            return Result.Fail(error)
        matches[0].at_bat_id = self.new_at_bat_id
        matches[0].batter_id = self.new_batter_id
        matches[0].batter_name = self.new_batter_name
        matches[0].is_patched = True
        return Result.Ok(data)


@dataclass
class PatchBrooksPitchFxPitcherId(Patch):
    bbref_game_id: str
    park_sv_id: str = field(repr=False)
    current_at_bat_id: str = field(repr=False)
    current_pitch_app_id: str = field(repr=False)
    current_pitcher_id: int
    current_pitcher_name: str = field(repr=False)
    new_at_bat_id: str = field(repr=False)
    new_pitch_app_id: str = field(repr=False)
    new_pitcher_id: int
    new_pitcher_name: str = field(repr=False)

    def __post_init__(self):
        self.data_set = DataSet.BROOKS_PITCHFX
        self.patch_id = "__patch_brooks_pitchfx_pitcher_id__"

    def apply(self, data):
        matches = [pfx for pfx in data if pfx.park_sv_id == self.park_sv_id]
        if not matches:
            error = (
                "Unable to locate the PitchFX reading identified in this patch: "
                f"pitch_app_id: {self.current_pitch_app_id}, park_sv_id: {self.park_sv_id}"
            )
            return Result.Fail(error)
        if len(matches) > 1:
            error = (
                "More than one PitchFX reading was found that matches this patch: "
                f"pitch_app_id: {self.current_pitch_app_id}, park_sv_id: {self.park_sv_id}"
            )
            return Result.Fail(error)
        matches[0].at_bat_id = self.new_at_bat_id
        matches[0].pitch_app_id = self.new_pitch_app_id
        matches[0].pitcher_id = self.new_pitcher_id
        matches[0].pitcher_name = self.new_pitcher_name
        matches[0].is_patched = True
        return Result.Ok(data)


@dataclass
class PatchBrooksPitchFxDeletePitchFx(Patch):
    bbref_game_id: str
    park_sv_id: str
    pitch_app_id: str

    def __post_init__(self):
        self.data_set = DataSet.BROOKS_PITCHFX
        self.patch_id = "__patch_brooks_pitchfx_delete_pitchfx__"

    def apply(self, data):
        pass



"""Investigate games where boxscore and PitchFx data could not be reconciled."""
import subprocess

from getch import pause

from vigorish.cli.components import print_heading, print_message
from vigorish.cli.menu_item import MenuItem
from vigorish.constants import EMOJIS
from vigorish.enums import AuditError
from vigorish.util.result import Result


class InvestigateScrapedDataFailures(MenuItem):
    def __init__(self, app, year, bbref_game_ids):
        super().__init__(app)
        self.year = year
        self.bbref_game_ids = bbref_game_ids
        self.menu_item_text = f"{AuditError.FAILED_TO_COMBINE} ({self.game_count} Games)"
        self.menu_item_emoji = EMOJIS.get("WEARY")

    @property
    def game_count(self):
        return len(self.bbref_game_ids)

    def launch(self):
        subprocess.run(["clear"])
        print_heading("This feature is not currently available", fg="bright_yellow")
        print_message("Sorry for any inconvenience, I promise to finish this soon!")
        pause(message="\nPress any key to return to the previous menu...")
        return Result.Ok(True)



"""Fix at bats with invalid PitchFX data."""
import subprocess
from copy import deepcopy

from getch import pause
from halo import Halo
from tabulate import tabulate

from vigorish.cli.components import (
    get_random_cli_color,
    get_random_dots_spinner,
    print_error,
    print_heading,
    print_message,
    print_success,
    select_game_prompt,
    user_options_prompt,
    yes_no_prompt,
)
from vigorish.cli.components.viewers import DictListTableViewer, DisplayTable, TableViewer
from vigorish.cli.menu_item import MenuItem
from vigorish.constants import EMOJIS, MENU_NUMBERS
from vigorish.enums import AuditError, PatchType
from vigorish.tasks import CombineScrapedDataTask, PatchAllInvalidPitchFxTask
from vigorish.tasks.patch_all_invalid_pfx import PatchInvalidPitchFxTask
from vigorish.util.result import Result
from vigorish.util.string_helpers import inning_number_to_string, validate_at_bat_id


class InvestigateInvalidPitchFx(MenuItem):
    def __init__(self, app, year, bbref_game_ids):
        super().__init__(app)
        self.combine_data = CombineScrapedDataTask(app)
        self.patch_all_invalid_pfx = PatchAllInvalidPitchFxTask(app)
        self.patch_invalid_pfx = PatchInvalidPitchFxTask(app)
        self.year = year
        self.bbref_game_ids = bbref_game_ids
        self.menu_item_text = f"{AuditError.INVALID_PITCHFX_DATA} ({self.game_count} Games)"
        self.menu_item_emoji = EMOJIS.get("UPSIDE_DOWN")

    @property
    def game_count(self):
        return len(self.bbref_game_ids)

    def launch(self):
        subprocess.run(["clear"])
        self.subscribe_to_events()
        result = self.select_patch_prompt()
        if result.failure:
            self.unsubscribe_from_events()
            return Result.Ok(True)
        selected_action = result.value
        if selected_action == "ALL":
            result = self.patch_all_invalid_pfx.execute(self.year)
            self.unsubscribe_from_events()
            if result.failure:
                return result
            patch_results_dict = result.value
            self.display_all_patch_results(**patch_results_dict)
            return Result.Ok(True)
        exit_single_game_menu = False
        while not exit_single_game_menu:
            result = select_game_prompt(self.bbref_game_ids)
            if result.failure:
                self.unsubscribe_from_events()
                return Result.Ok(True)
            self.game_id = result.value
            result = self.select_game_operation_prompt()
            if result.failure:
                exit_single_game_menu = True
                continue
            selected_action = result.value
            if selected_action == "INVESTIGATE":
                result = self.patch_invalid_pfx_single_game()
                if result.failure:
                    self.unsubscribe_from_events()
                    return result
            if selected_action == "RETRY":
                result = self.combine_scraped_data_for_game()
                if result.failure:
                    self.unsubscribe_from_events()
                    return result
        self.unsubscribe_from_events()
        return Result.Ok(True)

    def select_patch_prompt(self):
        prompt = (
            "You can attempt to patch all games with invald PitchFX data by selecting the first "
            "option below. Or, you can investigate each game in detail by selecting the second "
            "option."
        )
        choices = {
            f"{MENU_NUMBERS.get(1)}  Patch all Games": "ALL",
            f"{MENU_NUMBERS.get(2)}  Patch a Single Game": "ONE",
            f"{EMOJIS.get('BACK')} Return to Previous Menu": None,
        }
        return user_options_prompt(choices, prompt)

    def select_game_operation_prompt(self):
        prompt = "What would you like to do at this point?"
        choices = {
            f"{MENU_NUMBERS.get(1)}  Analyze Pitching Data": "INVESTIGATE",
            f"{MENU_NUMBERS.get(2)}  Attempt to Combine Data (Again)": "RETRY",
            f"{EMOJIS.get('BACK')} Return to Previous Menu": None,
        }
        return user_options_prompt(choices, prompt)

    def patch_invalid_pfx_single_game(self):
        result = self.patch_invalid_pfx.execute(self.game_id)
        if result.failure:
            header = f"Invalid PitchFX Data for {self.game_id}\n"
            subprocess.run(["clear"])
            print_message(header, wrap=False, bold=True, underline=True)
            print_message(result.error, fg="bright_yellow")
            pause(message="Press any key to continue...")
            return Result.Ok(True)
        subprocess.run(["clear"])
        self.display_invalid_pfx_details()
        if not self.prompt_user_create_patch_list():
            return Result.Ok(True)
        result = self.patch_invalid_pfx.match_missing_pfx_data()
        if result.failure:
            return result
        subprocess.run(["clear"])
        for result, matches in self.patch_invalid_pfx.match_results.items():
            if result == "success":
                for num, match_dict in enumerate(matches, start=1):
                    self.display_potential_data_match(num, len(matches), match_dict)
            if result == "no_matches":
                self.display_no_match_found(matches)
            if result == "many_matches":
                self.display_many_matches_found(matches)
        if "success" not in self.patch_invalid_pfx.match_results:
            header = f"Invalid PitchFX Data for {self.game_id}\n"
            message = (
                "Unable to identify missing data that matches the invalid PitchFX data for this "
                "game. You should inspect the combined data JSON file for this game and "
                "investigate the invalid data manually.\n"
            )
            subprocess.run(["clear"])
            print_message(header, wrap=False, bold=True, underline=True)
            print_message(message, fg="bright_yellow")
            pause(message="Press any key to continue...")
            return Result.Ok(True)
        result = self.patch_invalid_pfx.create_patch_list()
        if result.failure:
            return result
        if not self.patch_invalid_pfx.patch_list or not self.prompt_user_apply_patch_list():
            return Result.Ok(True)
        result = self.patch_invalid_pfx.apply_patch_list()
        if result.failure:
            return result
        self.patch_results = result.value
        print()
        if self.patch_results["fixed_all_errors"]:
            patch_result = f"PitchFX data for {self.game_id} is now completely reconciled (no errors of any type)!\n"
            print_success(patch_result)
        if self.patch_results["invalid_pfx"]:
            patch_result = f"{self.game_id} still contains invalid PitchFX data after applying the patch list.\n"
            print_error(patch_result)
        if self.patch_results["pfx_errors"]:
            patch_result = f"{self.game_id} still contains PitchFX data errors associated with valid at bats.\n"
            print_error(patch_result)
        pause(message="Press any key to continue...")
        subprocess.run(["clear"])
        if self.prompt_user_view_patched_data():
            self.display_patched_data_tables(**self.patch_results["patch_diff_report"])
        return Result.Ok()

    def display_invalid_pfx_details(self):
        invalid_pfx_error_data = []
        for pitch_app_dict in self.patch_invalid_pfx.invalid_pfx_map.values():
            for inning_id, inning_dict in pitch_app_dict.items():
                for at_bat_id, ab_dict in inning_dict.items():
                    ab_val = validate_at_bat_id(at_bat_id).value
                    ab_data = ab_dict["at_bat_data"]
                    pfx_audit = ab_data["at_bat_pitchfx_audit"]
                    pitch_numbers = sorted(pfx["ab_count"] for pfx in ab_data["pitchfx"])
                    pitch_count = max(pfx["ab_total"] for pfx in ab_data["pitchfx"])
                    error_data = {
                        "inning_id": inning_id[-5:],
                        "pitcher": f'{ab_data["pitcher_name"]} ({ab_val["pitcher_team"]})',
                        "batter": f'{ab_data["batter_name"]} ({ab_val["batter_team"]})',
                        "pitch_count": pitch_count,
                        "pfx_count": pfx_audit["pitch_count_pitchfx"],
                        "pfx_ab_numbers": pitch_numbers,
                    }
                    invalid_pfx_error_data.append(error_data)
        invalid_pfx_table = tabulate(invalid_pfx_error_data, headers="keys")
        self.summarize_pfx_errors(self.patch_invalid_pfx.invalid_pfx_map)
        print_message(invalid_pfx_table, wrap=False, fg="bright_magenta")

    def summarize_pfx_errors(self, pfx_error_map):
        at_bat_ids = [
            at_bat_id
            for pitch_app_dict in pfx_error_map.values()
            for inning_dict in pitch_app_dict.values()
            for at_bat_id in inning_dict.keys()
        ]
        pitch_app_ids = list(pfx_error_map.keys())
        at_bats_plural = "at bats" if len(at_bat_ids) > 1 else "at bat"
        pitch_apps_plural = "pitching appearances" if len(pitch_app_ids) > 1 else "pitching appearance"
        header = f"Invalid PitchFX Data for {self.game_id}\n"
        message = (
            f"{self.game_id} contains PitchFX data that does not belong to any at bat "
            f"that occurred during this game according to the baseball-reference.com boxscore "
            f"({len(at_bat_ids)} {at_bats_plural}, {len(pitch_app_ids)} total "
            f"{pitch_apps_plural}).\n"
        )
        print_message(header, wrap=False, bold=True, underline=True)
        print_message(message)

    def prompt_user_create_patch_list(self):
        print()
        prompt = (
            "\nInvalid PitchFX data is sometimes valid data where the pitcher or batter ID value "
            "is incorrect, while the inning, pitch count, and pitch numbers are correct. Would "
            "you like to check for at bats that match these criteria?"
        )
        return yes_no_prompt(prompt)

    def display_potential_data_match(self, num, total_matches, match_dict):
        self.describe_potential_data_match(num, total_matches, match_dict)
        match_copy = deepcopy(match_dict)
        match_copy["invalid_pfx"].pop("at_bat_id")
        match_copy["invalid_pfx"].pop("pitcher_id")
        match_copy["invalid_pfx"].pop("batter_id")
        match_copy["missing_pfx"].pop("at_bat_id")
        match_copy["missing_pfx"].pop("pitcher_id")
        match_copy["missing_pfx"].pop("batter_id")
        match_rows = [match_copy["invalid_pfx"], match_copy["missing_pfx"]]
        print_message(tabulate(match_rows, headers="keys"), wrap=False, fg="bright_green")
        match_dict["patch"] = self.prompt_user_create_patch(match_dict)

    def describe_potential_data_match(self, num, total, match_dict):
        invalid_pfx = match_dict["invalid_pfx"]
        missing_pfx = match_dict["missing_pfx"]
        at_bat_id = missing_pfx["at_bat_id"]
        inning_str = inning_number_to_string(invalid_pfx["inning_id"])
        pitcher_or_batter = "pitcher" if match_dict["patch_type"] == PatchType.CHANGE_BATTER_ID else "batter"
        match_heading = f"Match {num}/{total} (Patch Type: {match_dict['patch_type']}):"
        match_summary = []
        match_summary = (
            f"The invalid PitchFX data below for an at bat between {invalid_pfx['pitcher']} and "
            f"{invalid_pfx['batter']} in the {inning_str} is a potential "
            "match for missing data that actually occurred in the same inning between "
            f"{missing_pfx['pitcher']} and {missing_pfx['batter']} (at_bat_id: {at_bat_id}).\n"
        )
        match_details_heading = "The data used to make this potential match is given below:"
        match_details = [
            f"- Both at bats took place in the same inning ({missing_pfx['inning_id']}) and "
            f"involve the same {pitcher_or_batter}.\n",
            "- The number of pitches missing from the valid at bat "
            f"({len(missing_pfx['missing_pfx'])}) is the same (or less than) the total number of "
            f"pitches in the invalid at bat ({invalid_pfx['pitch_count']}).\n",
            f"- The pitch sequence numbers of the invalid at bat ({invalid_pfx['invalid_pfx']}) "
            "contain all of the pitch numbers that are missing from the valid at bat "
            f"({missing_pfx['missing_pfx']}).\n",
        ]

        subprocess.run(["clear"])
        print_heading(match_heading)
        print_message(match_summary)
        print_heading(match_details_heading, fg="bright_green")
        for message in match_details:
            print_message(message, fg="bright_green")

    def prompt_user_create_patch(self, match_dict):
        prompt = ""
        if match_dict["patch_type"] == PatchType.CHANGE_BATTER_ID:
            prompt = (
                "Change batter_id on all invalid_pfx data in the table above from "
                f'{self.get_batter_name_and_id(match_dict["invalid_pfx"])} to '
                f'{self.get_batter_name_and_id(match_dict["missing_pfx"])}'
            )
        if match_dict["patch_type"] == PatchType.CHANGE_PITCHER_ID:
            prompt = (
                "Change pitcher_id on all invalid_pfx data in the table above from "
                f'{self.get_pitcher_name_and_id(match_dict["invalid_pfx"])} to '
                f'{self.get_pitcher_name_and_id(match_dict["missing_pfx"])}'
            )
        prompt_heading = "\nWould you like to create a patch file to apply these changes?"
        print_heading(prompt_heading, fg="bright_yellow")
        return yes_no_prompt(prompt)

    def get_pitcher_name_and_id(self, at_bat_data):
        player_name = at_bat_data["pitcher"].split("(")[0].strip()
        player_id = at_bat_data["pitcher_id"]
        return f"{player_name} (MLB ID: {player_id})"

    def get_batter_name_and_id(self, at_bat_data):
        player_name = at_bat_data["batter"].split("(")[0].strip()
        player_id = at_bat_data["batter_id"]
        return f"{player_name} (MLB ID: {player_id})"

    def display_no_match_found(self, matches):
        for match_dict in matches:
            match_dict["invalid_pfx"].pop("at_bat_id")
            match_dict["invalid_pfx"].pop("pitcher_id")
            match_dict["invalid_pfx"].pop("batter_id")
        subprocess.run(["clear"])
        error = "No matching at bats were found for the invalid PitchFX data below:\n"
        print_message(error, fg="bright_red", bold=True)
        unmatched_rows = [match_dict["invalid_pfx"] for match_dict in matches]
        print_message(tabulate(unmatched_rows, headers="keys"), wrap=False)
        print()
        pause(message="Press any key to continue...")

    def display_many_matches_found(self, matches):
        for match_dict in matches:
            match_dict["invalid_pfx"].pop("at_bat_id")
            match_dict["invalid_pfx"].pop("pitcher_id")
            match_dict["invalid_pfx"].pop("batter_id")
            for match in match_dict["missing_pfx"]:
                match.pop("at_bat_id")
                match.pop("pitcher_id")
                match.pop("batter_id")
            subprocess.run(["clear"])
            error = "Multiple at bats were found for the invalid PitchFX data below (only one match is expected):\n"
            print_message(error, fg="bright_yellow", bold=True)
            all_rows = [match_dict["invalid_pfx"], *list(match_dict["missing_pfx"])]
            print_message(tabulate(all_rows, headers="keys"), wrap=False)
            print()
            pause(message="Press any key to continue...")

    def prompt_user_apply_patch_list(self):
        subprocess.run(["clear"])
        prompt = (
            f"\nA patch list for {self.game_id} was successfully created! Would you like to apply "
            "these changes and attempt to combine the scraped data after doing so?"
        )
        return yes_no_prompt(prompt)

    def combine_scraped_data_for_game(self):
        subprocess.run(["clear"])
        spinner = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
        spinner.text = f"Combining scraped data for {self.game_id}..."
        spinner.start()
        result = self.combine_data.execute(self.game_id)
        if (
            not result["gather_scraped_data_success"]
            or not result["combined_data_success"]
            or not result["update_pitch_apps_success"]
        ):
            spinner.fail(f"Failed to combine data for {self.game_id}!")
            pause(message="Press any key to continue...")
            return Result.Fail(result["error"])
        pfx_errors = result["results"]["pfx_errors"]
        fail_results = [
            pfx_errors.pop("pitchfx_error", {}),
            pfx_errors.pop("invalid_pitchfx", {}),
        ]
        if all(len(f) <= 0 for f in fail_results):
            spinner.succeed(f"All scraped data for {self.game_id} was successfully combined!")
            pause(message="Press any key to continue...")
            return Result.Ok()
        spinner.stop()
        subprocess.run(["clear"])
        total_pitch_apps = sum(len(f.keys()) for f in fail_results if f)
        pitch_apps_plural = "pitch appearances" if total_pitch_apps > 1 else "pitch appearance"
        total_at_bats = sum(len(at_bat_ids) for f in fail_results for at_bat_ids in f.values() if f)
        at_bats_plural = "at bats" if total_at_bats > 1 else "at bat"
        error_header = f"PitchFX data could not be reconciled for game: {self.game_id}\n"
        error_message = (
            f"{total_pitch_apps} {pitch_apps_plural} with data errors ({total_at_bats} total {at_bats_plural})\n"
        )
        print_message(error_header, wrap=False, fg="bright_red", bold=True, underline=True)
        print_message(error_message, fg="bright_red")
        if not self.prompt_user_investigate_failures():
            pause(message="Press any key to continue...")
            return Result.Ok()
        subprocess.run(["clear"])
        return self.patch_invalid_pfx_single_game()

    def prompt_user_investigate_failures(self):
        prompt = (
            "\nWould you like to analyze these results? In many cases, the batter/pitcher ID is "
            "incorrect and can be easily fixed by applying a patch file."
        )
        return yes_no_prompt(prompt)

    def prompt_user_view_patched_data(self):
        prompt = "\nWould you like to see a report detailing the changes that were made by applying the patch list?"
        return yes_no_prompt(prompt)

    def display_patched_data_tables(self, boxscore_changes, pitch_stat_changes_dict, pitch_stats_after):
        table_list = []
        total_tables = len(pitch_stat_changes_dict) + 1
        boxscore_heading = f"Table 1/{total_tables}: All Pitch Data for {self.game_id}\n"
        boxscore_table = DisplayTable(boxscore_changes, boxscore_heading)
        table_list.append(boxscore_table)
        for num, (pitch_app_id, pitch_stat_changes) in enumerate(pitch_stat_changes_dict.items(), start=2):
            pitch_app_header = f"Table {num}/{total_tables}: Pitch Appearance {pitch_app_id}\n"
            pitch_app_message = self.get_pitch_app_stat_line(pitch_stats_after[pitch_app_id])
            pitch_app_table = DisplayTable(pitch_stat_changes, pitch_app_header, pitch_app_message)
            table_list.append(pitch_app_table)
        table_viewer = TableViewer(
            table_list=table_list,
            prompt="Press Enter to return to previous menu",
            confirm_only=True,
            table_color="bright_cyan",
            heading_color="bright_yellow",
            message_color=None,
        )
        table_viewer.launch()

    def get_pitch_app_stat_line(self, pitch_stats):
        name = pitch_stats["pitcher_name"]
        line = pitch_stats["bbref_data"]
        game_score = f" (GS: {line['game_score']})\n" if line["game_score"] > 0 else "\n"
        earned_runs = (
            f"{line['earned_runs']} ER"
            if line["earned_runs"] == line["runs"]
            else f"{line['runs']} R ({line['earned_runs']} ER)"
        )
        return (
            f"{name}: {line['innings_pitched']} IP, {earned_runs}, {line['hits']} H, "
            f"{line['strikeouts']} K, {line['bases_on_balls']} BB{game_score}"
        )

    def display_all_patch_results(self, all_patch_results, successful_change, invalid_pfx_change):
        col_headers = ["BBRef Game ID", "Status"]
        game_results = [
            [game_id, self.get_scrape_status(patch_results)] for game_id, patch_results in all_patch_results.items()
        ]
        game_results.sort(key=lambda x: (x[1], x[0]))
        row_data = [dict(zip(col_headers, row)) for row in game_results]
        heading = f"Operation: Patch all Invalid PitchFX Data for MLB {self.year}"
        message = self.get_patch_results_message(all_patch_results, game_results)
        table_viewer = DictListTableViewer(
            row_data,
            prompt="Press Enter to return to previous menu",
            confirm_only=True,
            heading=heading,
            heading_color="bright_yellow",
            message=message,
            table_color="bright_cyan",
        )
        table_viewer.launch()

    def get_scrape_status(self, patch_results):
        if not patch_results["created_patch_list"]:
            return "failed to create patch list"
        if patch_results["fixed_all_errors"]:
            return "fixed all errors"
        if patch_results["invalid_pfx"] and patch_results["pfx_errors"]:
            return "invalid and valid pfx errors remain"
        if patch_results["invalid_pfx"]:
            return "invalid pfx errors remain"
        if patch_results["pfx_errors"]:
            return "valid pfx errors remain"

    def get_patch_results_message(self, all_patch_results, game_results):
        patch_list_created_count = self.get_patch_list_created_count(all_patch_results)
        patch_list_failed_count = self.get_patch_list_failed_count(all_patch_results)
        fixed_all_count = self.get_fixed_all_count(all_patch_results)
        remaining_errors = self.get_remaining_error_count(all_patch_results)
        messages = [f"Processed {len(game_results)} total games"]
        if patch_list_failed_count:
            messages.append(f"  - Failed to create a patch list for {patch_list_failed_count} games")
        if patch_list_created_count:
            messages.append(f"  - Created a patch list for {patch_list_created_count} games")
        if fixed_all_count:
            messages.append(f"  - Eliminated all PitchFX errors from {fixed_all_count} games")
        if remaining_errors:
            messages.append(f"  - PitchFX data errors still exist in {remaining_errors} games")
        return "\n".join(messages)

    def get_patch_list_created_count(self, all_patch_results):
        return len([r for r in all_patch_results.values() if "created_patch_list" in r and r["created_patch_list"]])

    def get_patch_list_failed_count(self, all_patch_results):
        return len([r for r in all_patch_results.values() if "created_patch_list" in r and not r["created_patch_list"]])

    def get_fixed_all_count(self, all_patch_results):
        return len([r for r in all_patch_results.values() if "fixed_all_errors" in r and r["fixed_all_errors"]])

    def get_invalid_pfx_count(self, all_patch_results):
        return len([r for r in all_patch_results.values() if "invalid_pfx" in r and r["invalid_pfx"]])

    def get_pfx_error_count(self, all_patch_results):
        return len([r for r in all_patch_results.values() if "pfx_errors" in r and r["pfx_errors"]])

    def get_remaining_error_count(self, all_patch_results):
        return (
            self.get_patch_list_failed_count(all_patch_results)
            + self.get_invalid_pfx_count(all_patch_results)
            + self.get_pfx_error_count(all_patch_results)
        )

    def error_occurred(self, error_message):
        print_error(error_message)

    def patch_all_invalid_pitchfx_started(self):
        pass

    def patch_all_invalid_pitchfx_complete(self):
        pass

    def create_invalid_pfx_map_start(self):
        pass

    def create_invalid_pfx_map_complete(self, invalid_pfx_map):
        self.invalid_pfx_map = invalid_pfx_map

    def match_missing_pfx_data_start(self):
        pass

    def match_missing_pfx_data_complete(self, match_results):
        self.match_results = match_results

    def create_patch_list_start(self):
        pass

    def create_patch_list_complete(self, patch_list):
        self.patch_list = patch_list

    def combine_scraped_data_start(self):
        pass

    def combine_scraped_data_complete(self, patch_results):
        self.patch_results = patch_results

    def subscribe_to_events(self):
        self.patch_all_invalid_pfx.events.patch_all_invalid_pitchfx_started += self.patch_all_invalid_pitchfx_started
        self.patch_all_invalid_pfx.events.patch_all_invalid_pitchfx_complete += self.patch_all_invalid_pitchfx_complete
        self.patch_all_invalid_pfx.events.create_invalid_pfx_map_start += self.create_invalid_pfx_map_start
        self.patch_all_invalid_pfx.events.create_invalid_pfx_map_complete += self.create_invalid_pfx_map_complete
        self.patch_all_invalid_pfx.events.error_occurred += self.error_occurred
        self.patch_all_invalid_pfx.events.match_missing_pfx_data_start += self.match_missing_pfx_data_start
        self.patch_all_invalid_pfx.events.match_missing_pfx_data_complete += self.match_missing_pfx_data_complete
        self.patch_all_invalid_pfx.events.create_patch_list_start += self.create_patch_list_start
        self.patch_all_invalid_pfx.events.create_patch_list_complete += self.create_patch_list_complete
        self.patch_all_invalid_pfx.events.combine_scraped_data_start += self.combine_scraped_data_start
        self.patch_all_invalid_pfx.events.combine_scraped_data_complete += self.combine_scraped_data_complete

    def unsubscribe_from_events(self):
        self.patch_all_invalid_pfx.events.create_invalid_pfx_map_start -= self.create_invalid_pfx_map_start
        self.patch_all_invalid_pfx.events.create_invalid_pfx_map_complete -= self.create_invalid_pfx_map_complete
        self.patch_all_invalid_pfx.events.patch_all_invalid_pitchfx_started -= self.patch_all_invalid_pitchfx_started
        self.patch_all_invalid_pfx.events.patch_all_invalid_pitchfx_complete -= self.patch_all_invalid_pitchfx_complete
        self.patch_all_invalid_pfx.events.error_occurred -= self.error_occurred
        self.patch_all_invalid_pfx.events.match_missing_pfx_data_start -= self.match_missing_pfx_data_start
        self.patch_all_invalid_pfx.events.match_missing_pfx_data_complete -= self.match_missing_pfx_data_complete
        self.patch_all_invalid_pfx.events.create_patch_list_start -= self.create_patch_list_start
        self.patch_all_invalid_pfx.events.create_patch_list_complete -= self.create_patch_list_complete
        self.patch_all_invalid_pfx.events.combine_scraped_data_start -= self.combine_scraped_data_start
        self.patch_all_invalid_pfx.events.combine_scraped_data_complete -= self.combine_scraped_data_complete



from tabulate import tabulate

from vigorish.cli.components.viewers.display_table import DisplayTable
from vigorish.cli.components.viewers.table_viewer import TableViewer
from vigorish.util.list_helpers import make_chunked_list


class DictListTableViewer(TableViewer):
    def __init__(
        self,
        dict_list,
        prompt=None,
        confirm_only=False,
        yes_choice="YES",
        no_choice="NO",
        rows_per_page=10,
        row_numbers=False,
        heading=None,
        heading_color=None,
        message=None,
        message_color=None,
        table_color=None,
    ):
        super().__init__(
            table_list=self.create_table_list(dict_list, rows_per_page, row_numbers, heading, message),
            prompt=prompt,
            confirm_only=confirm_only,
            yes_choice=yes_choice,
            no_choice=no_choice,
            heading_color=heading_color,
            message_color=message_color,
            table_color=table_color,
        )

    def create_table_list(self, dict_list, rows_per_page, row_numbers, heading, message):
        table_dicts = self.create_table_dicts(dict_list, row_numbers)
        table_chunks = make_chunked_list(table_dicts, chunk_size=rows_per_page)
        return [DisplayTable(tabulate(d, headers="keys"), heading, message) for d in table_chunks]

    def create_table_dicts(self, dict_list, row_numbers):
        if not dict_list:
            return ([], [])
        if not row_numbers:
            return dict_list
        col_names = list(dict_list[0].keys())
        table_rows = [list(row_dict.values()) for row_dict in dict_list]
        for num, row in enumerate(table_rows, start=1):
            row.insert(0, num)
        col_names.insert(0, "#")
        return [dict(zip(col_names, row)) for row in table_rows if row != [""]]



"""Menu that allows the user to execute incomplete jobs and view status of completed jobs."""
from vigorish.cli.menu import Menu
from vigorish.cli.menu_items.job_details import JobDetails
from vigorish.cli.menu_items.return_to_parent import ReturnToParent
from vigorish.constants import MENU_NUMBERS


class JobsMenu(Menu):
    def __init__(self, app, jobs, job_status, menu_item_number):
        super().__init__(app)
        self.jobs = jobs
        self.job_status = job_status
        self.menu_text = f"Status = {self.job_status.name.title()} ({len(jobs)} total jobs)"
        self.menu_item_text = f" {self.job_status.name.title()} Jobs ({len(jobs)})"
        self.menu_item_emoji = MENU_NUMBERS.get(menu_item_number, str(menu_item_number))

    def populate_menu_items(self):
        self.menu_items = [
            JobDetails(self.app, job, menu_item_number) for menu_item_number, job in enumerate(self.jobs, start=1)
        ]
        self.menu_items.append(ReturnToParent(self.app, "Return to All Jobs Menu "))
        if len(self.jobs) > 8:
            self.menu_items.insert(0, ReturnToParent(self.app, "Return to All Jobs Menu"))



from pathlib import Path

from events import Events

import vigorish.database as db
from vigorish.enums import DataSet, VigFile
from vigorish.status.update_status_bbref_boxscores import update_status_bbref_boxscore_list
from vigorish.status.update_status_bbref_games_for_date import update_bbref_games_for_date_list
from vigorish.status.update_status_brooks_games_for_date import (
    update_status_brooks_games_for_date_list,
)
from vigorish.status.update_status_brooks_pitch_logs import (
    update_status_brooks_pitch_logs_for_game_list,
)
from vigorish.status.update_status_brooks_pitchfx import update_status_brooks_pitchfx_log_list
from vigorish.status.update_status_combined_data import update_status_combined_data_list
from vigorish.tasks.base import Task
from vigorish.util.result import Result


class ImportScrapedDataTask(Task):
    def __init__(self, app):
        super().__init__(app)
        self.events = Events(
            (
                "error_occurred",
                "no_scraped_data_found",
                "search_local_files_start",
                "search_local_files_complete",
                "import_scraped_data_start",
                "import_scraped_data_complete",
                "import_scraped_data_for_year_start",
                "import_scraped_data_for_year_complete",
                "import_scraped_data_set_start",
                "import_scraped_data_set_complete",
                "import_combined_game_data_start",
                "import_combined_game_data_complete",
            )
        )

    def execute(self, overwrite_existing=False):
        mlb_season_map = self._get_mlb_season_scraped_data_map()
        if not any(season["has_data"] for season in mlb_season_map.values()):
            self.events.no_scraped_data_found()
            return Result.Ok()
        self._remove_invalid_pitchfx_logs(mlb_season_map)
        return self._import_scraped_data(mlb_season_map, overwrite_existing)

    def _get_mlb_season_scraped_data_map(self):
        self.events.search_local_files_start()
        all_mlb_seasons = db.Season.get_all_regular_seasons(self.db_session)
        return {
            season.year: {
                "has_data": self._local_folder_has_parsed_data_for_season(season.year),
                "season": season,
            }
            for season in all_mlb_seasons
        }

    def _local_folder_has_parsed_data_for_season(self, year):
        for data_set in DataSet:
            scraped_ids = self.scraped_data.get_scraped_ids_from_local_folder(VigFile.PARSED_JSON, data_set, year)
            if scraped_ids:
                return True
        return False

    def _remove_invalid_pitchfx_logs(self, mlb_season_map):
        for season in mlb_season_map.values():
            if season["has_data"]:
                invalid_ids = self._get_invalid_pitchfx_log_ids_for_season(season["season"])
                if invalid_ids:
                    self._delete_invalid_pitchfx_json_files(season["season"], invalid_ids)
        self.events.search_local_files_complete()

    def _get_invalid_pitchfx_log_ids_for_season(self, season):
        valid_ids = []
        for game_date in season.get_date_range():
            games_for_date = self.scraped_data.get_brooks_games_for_date(game_date)
            if not games_for_date:
                continue
            valid_ids.extend(games_for_date.all_pitch_app_ids_for_date)
        local_ids = self.scraped_data.get_scraped_ids_from_local_folder(
            VigFile.PARSED_JSON, DataSet.BROOKS_PITCHFX, season.year
        )
        return list(set(local_ids) - set(valid_ids))

    def _delete_invalid_pitchfx_json_files(self, season, invalid_ids):
        pfx_folderpath = self.scraped_data.get_local_folderpath(
            VigFile.PARSED_JSON, DataSet.BROOKS_PITCHFX, season.year
        )
        for pitch_app_id in invalid_ids:
            invalid_pitchfx_log = Path(pfx_folderpath).joinpath(f"{pitch_app_id}.json")
            if invalid_pitchfx_log.exists():
                invalid_pitchfx_log.unlink()

    def _import_scraped_data(self, mlb_season_map, overwrite_existing):
        self.events.import_scraped_data_start()
        error_dict = {}
        for season in mlb_season_map.values():
            if not season["has_data"]:
                continue
            error_dict[season["season"].year] = []
            result = self._update_all_data_sets(season["season"], overwrite_existing)
            if result.failure:
                error_dict[season["season"].year].append(result.error)
            result = self._update_combined_data(season["season"], overwrite_existing)
            if result.failure:
                error_dict[season["season"].year].append(result.error)
        self.events.import_scraped_data_complete()
        error_dict = self._check_error_dict(error_dict)
        return Result.Ok() if not error_dict else Result.Fail(self._create_error_message_for_task(error_dict))

    def _check_error_dict(self, error_dict):
        for year in list(error_dict.keys())[:]:
            if not error_dict[year]:
                del error_dict[year]

    def _create_error_message_for_task(self, error_dict):
        error_list = [f"{year}: " + "\n".join(errors) for year, errors in error_dict.items()]
        return "\n".join(error_list)

    def _update_all_data_sets(self, season, overwrite_existing=False):
        error_dict = {}
        self.events.import_scraped_data_for_year_start(season.year)
        for data_set in DataSet:
            result = self._update_data_set(data_set, season, overwrite_existing)
            if result.failure:
                error_dict[data_set] = result.error
                continue
        self.events.import_scraped_data_for_year_complete(season.year)
        return Result.Ok() if not error_dict else Result.Fail(self._create_error_message_for_season(error_dict))

    def _create_error_message_for_season(self, error_dict):
        error_list = [f"{data_set.name}: {error_ids}" for data_set, error_ids in error_dict.items()]
        return "\n".join(error_list)

    def _update_data_set(self, data_set, season, overwrite_existing=False):
        self.events.import_scraped_data_set_start(data_set, season.year)
        scraped_ids = self.scraped_data.get_scraped_ids_from_local_folder(VigFile.PARSED_JSON, data_set, season.year)
        if not scraped_ids:
            return Result.Ok()
        if not overwrite_existing:
            existing_scraped_ids = self.scraped_data.get_scraped_ids_from_database(data_set, season)
            scraped_ids = list(set(scraped_ids) - set(existing_scraped_ids))
        result = self._update_status_for_data_set(data_set, scraped_ids)
        self.db_session.commit()
        self.events.import_scraped_data_set_complete(data_set, season.year)
        return result

    def _update_combined_data(self, season, overwrite_existing=False):
        self.events.import_combined_game_data_start(season.year)
        combined_ids = self.scraped_data.get_scraped_ids_from_local_folder(
            VigFile.COMBINED_GAME_DATA, DataSet.ALL, season.year
        )
        if not combined_ids:
            self.events.import_combined_game_data_complete(season.year)
            return Result.Ok()
        if not overwrite_existing:
            season_report = self.app.audit_report.get(season.year, {})
            if season_report:
                existing_combined_ids = (
                    season_report["successful"] + season_report["pfx_error"] + season_report["invalid_pfx"]
                )
                combined_ids = list(set(combined_ids) - set(existing_combined_ids))
        result = update_status_combined_data_list(self.scraped_data, self.db_session, combined_ids)
        self.events.import_combined_game_data_complete(season.year)
        return result

    def _update_status_for_data_set(self, data_set, scraped_ids):
        update_status_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: update_status_brooks_games_for_date_list,
            DataSet.BBREF_GAMES_FOR_DATE: update_bbref_games_for_date_list,
            DataSet.BBREF_BOXSCORES: update_status_bbref_boxscore_list,
            DataSet.BROOKS_PITCH_LOGS: update_status_brooks_pitch_logs_for_game_list,
            DataSet.BROOKS_PITCHFX: update_status_brooks_pitchfx_log_list,
        }
        return update_status_dict[data_set](self.scraped_data, self.db_session, scraped_ids)



"""Database file and connection details."""
# flake8: noqa
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()


from vigorish.models import (
    Assoc_Player_Team,
    BatStats,
    BatStatsCsvRow,
    DateScrapeStatus,
    DateScrapeStatusCsvRow,
    GameScrapeStatus,
    GameScrapeStatusCsvRow,
    PitchAppScrapeStatus,
    PitchAppScrapeStatusCsvRow,
    PitchFx,
    PitchFxCsvRow,
    PitchStats,
    PitchStatsCsvRow,
    PitchTypePercentile,
    Player,
    PlayerId,
    ScrapeError,
    ScrapeJob,
    Season,
    Team,
    TimeBetweenPitches,
)
from vigorish.views import (
    BatStats_All_View,
    BatStats_By_Opp_Team_View,
    BatStats_By_Opp_Team_Year_View,
    BatStats_By_Team_View,
    BatStats_By_Team_Year_View,
    BatStats_By_Year_View,
    Batter_PitchFx_All_View,
    Batter_PitchFx_By_Year_View,
    Batter_PitchFx_For_Game_All_View,
    Batter_PitchFx_For_Game_vs_LHP_as_LHB_View,
    Batter_PitchFx_For_Game_vs_LHP_as_RHB_View,
    Batter_PitchFx_For_Game_vs_RHP_as_LHB_View,
    Batter_PitchFx_For_Game_vs_RHP_as_RHB_View,
    Batter_PitchFx_vs_LHP_as_LHB_Career_View,
    Batter_PitchFx_vs_LHP_as_RHB_Career_View,
    Batter_PitchFx_vs_RHP_as_LHB_Career_View,
    Batter_PitchFx_vs_RHP_as_RHB_Career_View,
    Batter_PitchType_All_View,
    Batter_PitchType_By_Year_View,
    Batter_PitchType_For_Game_All_View,
    Batter_PitchType_For_Game_vs_LHP_as_LHB_View,
    Batter_PitchType_For_Game_vs_LHP_as_RHB_View,
    Batter_PitchType_For_Game_vs_RHP_as_LHB_View,
    Batter_PitchType_For_Game_vs_RHP_as_RHB_View,
    Batter_PitchType_vs_LHP_as_LHB_Career_View,
    Batter_PitchType_vs_LHP_as_RHB_Career_View,
    Batter_PitchType_vs_RHP_as_LHB_Career_View,
    Batter_PitchType_vs_RHP_as_RHB_Career_View,
    Date_PitchApp_View,
    Game_PitchApp_View,
    Pitcher_PitchFx_All_By_Year_View,
    Pitcher_PitchFx_All_View,
    Pitcher_PitchFx_For_Game_All_View,
    Pitcher_PitchFx_For_Game_vs_LHB_View,
    Pitcher_PitchFx_For_Game_vs_RHB_View,
    Pitcher_PitchFx_vs_LHB_By_Year_View,
    Pitcher_PitchFx_vs_LHB_View,
    Pitcher_PitchFx_vs_RHB_By_Year_View,
    Pitcher_PitchFx_vs_RHB_View,
    Pitcher_PitchType_All_By_Year_View,
    Pitcher_PitchType_All_View,
    Pitcher_PitchType_For_Game_All_View,
    Pitcher_PitchType_For_Game_vs_LHB_View,
    Pitcher_PitchType_For_Game_vs_RHB_View,
    Pitcher_PitchType_vs_LHB_By_Year_View,
    Pitcher_PitchType_vs_LHB_View,
    Pitcher_PitchType_vs_RHB_By_Year_View,
    Pitcher_PitchType_vs_RHB_View,
    PitchStats_All_View,
    PitchStats_By_Opp_Team_By_Year_View,
    PitchStats_By_Opp_Team_View,
    PitchStats_By_Team_By_Year_View,
    PitchStats_By_Team_View,
    PitchStats_By_Year_View,
    PitchStats_RP_View,
    PitchStats_SP_View,
    PitchType_All_View,
    PitchType_Left_View,
    PitchType_Right_View,
    Season_Date_View,
    Season_Game_PitchApp_View,
    Season_Game_View,
    Season_PitchApp_View,
    Team_BatStats_By_BatOrder_By_Player_By_Year,
    Team_BatStats_By_BatOrder_By_Year,
    Team_BatStats_By_DefPosition_By_Player_By_Year,
    Team_BatStats_By_DefPosition_By_Year,
    Team_BatStats_By_Player_By_Year_View,
    Team_BatStats_By_Year_View,
    Team_BatStats_For_Starters_By_Player_By_Year,
    Team_BatStats_For_Starters_By_Year,
    Team_BatStats_For_Subs_By_Player_By_Year,
    Team_BatStats_For_Subs_By_Year,
    Team_PitchStats_By_Player_By_Year_View,
    Team_PitchStats_By_Year_View,
    Team_PitchStats_RP_By_Player_By_Year_View,
    Team_PitchStats_RP_By_Year_View,
    Team_PitchStats_SP_By_Player_By_Year_View,
    Team_PitchStats_SP_By_Year_View,
)



"""Populate reference tables with data."""
import vigorish.database as db
from vigorish.util.result import Result


def populate_base_tables(db_session):
    """Populate reference tables with data."""
    result = add_time_between_pitches(db_session)
    if result.failure:
        return result
    db_session.commit()
    return Result.Ok()


def add_time_between_pitches(db_session):
    result = db.TimeBetweenPitches.from_calc_results(db_session, get_avg_pitch_times())
    return Result.Ok() if result else Result.Fail("failed to populate time_between_pitches")


def get_avg_pitch_times():
    return {
        "time_between_pitches": {
            "total": 16478993,
            "count": 692284,
            "avg": 23.8,
            "max": 48.0,
            "min": 3.0,
        },
        "time_between_at_bats": {
            "total": 7118617,
            "count": 165166,
            "avg": 43.1,
            "max": 79.0,
            "min": 28.0,
        },
        "time_between_innings": {
            "total": 8432869,
            "count": 52589,
            "avg": 160.4,
            "max": 306.0,
            "min": 131.0,
        },
    }



import click
from dateutil import parser as date_parser

import vigorish.database as db
from vigorish.enums import DataSet, VigFile
from vigorish.util.regex import JOB_NAME_REGEX


class DateString(click.ParamType):
    name = "date-string"

    def convert(self, value, param, ctx):
        try:
            return date_parser.parse(value)
        except Exception:
            error = (
                f'"{value}" could not be parsed as a valid date. You can '
                "use any format recognized by dateutil.parser. For example, "
                "all of the strings below are valid ways to represent the "
                "same date\n"
                '"2018-5-13" -or- "05/13/2018" -or- "May 13 2018"'
            )
            self.fail(error, param, ctx)


class DataSetName(click.ParamType):
    name = "data-set"

    def convert(self, value, param, ctx):
        if value is None:
            return DataSet.NONE
        if value.upper() == "ALL":
            return DataSet.ALL
        if param.multiple:
            value_list = [s.strip() for s in value.split(",")]
            data_set_list = [(DataSet.from_str(val), val) for val in value_list]
            bad_values = list(filter(lambda x: x[0] == DataSet.NONE, data_set_list))
            if bad_values:
                bad_val_plural = "values" if len(bad_values) > 1 else "value"
                all_val_plural = "values" if len(value_list) > 1 else "value"
                is_plural = "are not valid data set names" if len(bad_values) > 1 else "is not a valid data set name"
                error = (
                    f"{len(bad_values)} {bad_val_plural} (of {len(value_list)} total {all_val_plural} "
                    f'provided) {is_plural}: {", ".join(v[1] for v in bad_values)}\n'
                    f'Valid data set names are: {", ".join(ds.name for ds in DataSet)}, ALL'
                )
                self.fail(error, param, ctx)
            return [ds[0] for ds in data_set_list]
        else:
            data_set = DataSet.from_str(value)
            if data_set:
                return data_set
            error = (
                f'"{value}" is not a valid name for a data set. Valid names are: '
                f'{", ".join(ds.name for ds in DataSet)}, ALL'
            )
            self.fail(error, param, ctx)


class FileTypeName(click.ParamType):
    name = "file-type"

    def convert(self, value, param, ctx):
        if value is None:
            return VigFile.NONE
        if value.upper() == "ALL":
            return VigFile.ALL
        if param.multiple:
            value_list = [s.strip() for s in value.split(",")]
            file_type_list = [(VigFile.from_str(val), val) for val in value_list]
            bad_values = list(filter(lambda x: x[0] == VigFile.NONE, file_type_list))
            if bad_values:
                bad_val_plural = "values" if len(bad_values) > 1 else "value"
                all_val_plural = "values" if len(value_list) > 1 else "value"
                is_plural = "are not valid file type names" if len(bad_values) > 1 else "is not a valid file type name"
                error = (
                    f"{len(bad_values)} {bad_val_plural} (of {len(value_list)} total {all_val_plural} "
                    f'provided) {is_plural}: {", ".join(v[1] for v in bad_values)}\n'
                    f'Valid file type names are: {", ".join(f.name for f in VigFile)}, ALL'
                )
                self.fail(error, param, ctx)
            return [f[0] for f in file_type_list]
        else:
            file_type = VigFile.from_str(value)
            if file_type:
                return file_type
            error = (
                f'"{value}" is not a valid file type name. Valid names are: '
                f'{", ".join(f.name for f in VigFile)}, ALL'
            )
            self.fail(error, param, ctx)


class MlbSeason(click.ParamType):
    name = "year-number"

    def convert(self, value, param, ctx):
        db_session = ctx.obj.db_session
        try:
            year = int(value)
        except Exception:
            error = f'Unable to parse "{value}" as an integer.'
            self.fail(error, param, ctx)

        try:
            season = db.Season.find_by_year(db_session, year)
            if season:
                return season.year
            valid_years = [s.year for s in db.Season.get_all_regular_seasons(db_session)]
            year_min = min(sorted(valid_years))
            year_max = max(sorted(valid_years))
            error = (
                f'"{value}" is not a valid value. Only MLB Seasons in '
                f"the range {year_min}-{year_max} are supported in this "
                "version of vig."
            )
            self.fail(error, param, ctx)
        except Exception as ex:
            self.fail(repr(ex), param, ctx)


class JobName(click.ParamType):
    name = "job-name"

    def convert(self, value, param, ctx):
        if not value:
            return value
        if JOB_NAME_REGEX.match(value):
            return value
        error = (
            f"'{value}' contains one or more invalid characters. Job name must "
            "contain only letters, numbers, hyphen and underscore characters."
        )
        self.fail(error, param, ctx)



"""Pitch logs and pitchfx URLs for a single game."""
import json
from dataclasses import dataclass, field
from typing import List

from vigorish.scrape.brooks_pitch_logs.models.pitch_log import BrooksPitchLog
from vigorish.util.list_helpers import as_dict_list


@dataclass
class BrooksPitchLogsForGame:
    """Pitch logs and pitchfx URLs for a single game."""

    bb_game_id: str = ""
    bbref_game_id: str = ""
    pitch_log_count: str = ""
    pitch_logs: List[BrooksPitchLog] = field(default_factory=list)

    def as_dict(self):
        """Convert pitch logs for game to a dictionary."""
        return {
            "__brooks_pitch_logs_for_game__": True,
            "bb_game_id": self.bb_game_id,
            "bbref_game_id": self.bbref_game_id,
            "pitch_log_count": int(self.pitch_log_count),
            "pitch_logs": as_dict_list(self.pitch_logs),
        }

    def as_json(self):
        """Convert pitch logs for game to JSON."""
        return json.dumps(self.as_dict(), indent=2)



import vigorish.database as db
from vigorish.enums import DataSet
from vigorish.util.dt_format_strings import DATE_ONLY, DATE_ONLY_TABLE_ID
from vigorish.util.result import Result


def update_brooks_games_for_date_single_date(db_session, season, games_for_date):
    game_date = games_for_date.game_date
    result = update_date_status_records(db_session, games_for_date, game_date)
    if result.failure:
        return result
    result = update_game_status_records(db_session, games_for_date)
    if result.failure:
        return result
    db_session.commit()
    return Result.Ok()


def update_status_brooks_games_for_date_list(scraped_data, db_session, scraped_brooks_dates, apply_patch_list=True):
    season = None
    for game_date in scraped_brooks_dates:
        if not season:
            season = db.Season.find_by_year(db_session, game_date.year)
        games_for_date = scraped_data.get_brooks_games_for_date(game_date, apply_patch_list)
        if not games_for_date:
            date_str = game_date.strftime(DATE_ONLY_TABLE_ID)
            error = f"Failed to retrieve {DataSet.BROOKS_GAMES_FOR_DATE} (URL ID: {date_str})"
            return Result.Fail(error)
        result = update_date_status_records(db_session, games_for_date, game_date)
        if result.failure:
            return result
        result = update_game_status_records(db_session, games_for_date)
        if result.failure:
            return result
        db_session.commit()
    return Result.Ok()


def update_date_status_records(db_session, games_for_date, game_date):
    try:
        date_id = game_date.strftime(DATE_ONLY_TABLE_ID)
        date_status = db_session.query(db.DateScrapeStatus).get(date_id)
        if not date_status:
            date_str = games_for_date.game_date.strftime(DATE_ONLY)
            error = f"scrape_status_date does not contain an entry for date: {date_str}"
            return Result.Fail(error)
        date_status.scraped_daily_dash_brooks = 1
        date_status.game_count_brooks = games_for_date.game_count
        return Result.Ok()
    except Exception as e:
        return Result.Fail(f"Error: {repr(e)}")


def update_game_status_records(db_session, games_for_date):
    for game_info in games_for_date.games:
        if not game_info.pitcher_appearance_count:
            continue
        try:
            bbref_game_id = game_info.bbref_game_id
            game_status = db.GameScrapeStatus.find_by_bbref_game_id(db_session, bbref_game_id)
            if not game_status:
                game_status = create_game_status_record(db_session, games_for_date, bbref_game_id)
            game_status.bb_game_id = game_info.bb_game_id
            game_status.game_time_hour = game_info.game_time_hour
            game_status.game_time_minute = game_info.game_time_minute
            game_status.game_time_zone = game_info.time_zone_name
            game_status.pitch_app_count_brooks = game_info.pitcher_appearance_count
        except Exception as e:
            return Result.Fail(f"Error: {repr(e)}")
    return Result.Ok()


def create_game_status_record(db_session, games_for_date, bbref_game_id):
    game_date = games_for_date.game_date
    date_status = db.DateScrapeStatus.find_by_date(db_session, game_date)
    date_status.scraped_daily_dash_brooks = 1
    date_status.game_count_brooks = games_for_date.game_count
    game_status = db.GameScrapeStatus()
    game_status.game_date = game_date
    game_status.bbref_game_id = bbref_game_id
    game_status.scrape_status_date_id = date_status.id
    game_status.season_id = date_status.season_id
    db_session.add(game_status)
    return game_status



"""Values used to identify a single player in various websites and databases."""
from sqlalchemy import Column, ForeignKey, Integer, String

import vigorish.database as db
from vigorish.util.string_helpers import validate_at_bat_id


class PlayerId(db.Base):
    """Values used to identify a single player in various websites and databases."""

    __tablename__ = "player_id"
    mlb_id = Column(Integer, primary_key=True)
    mlb_name = Column(String)
    bbref_id = Column(String, index=True)
    bbref_name = Column(String)
    db_player_id = Column(Integer, ForeignKey("player.id"))

    def __repr__(self):
        return f"<PlayerId bbref_id={self.bbref_id}, mlb_id={self.mlb_id}>"

    @classmethod
    def find_by_bbref_id(cls, db_session, bbref_id):
        return db_session.query(cls).filter_by(bbref_id=bbref_id).first()

    @classmethod
    def find_by_mlb_id(cls, db_session, mlb_id) -> "PlayerId":
        return db_session.query(cls).filter_by(mlb_id=mlb_id).first()

    @classmethod
    def get_player_ids_from_at_bat_id(cls, db_session, at_bat_id):
        at_bat_dict = validate_at_bat_id(at_bat_id).value
        pitcher_id_mlb = int(at_bat_dict["pitcher_mlb_id"])
        batter_id_mlb = int(at_bat_dict["batter_mlb_id"])
        pitcher_id = cls.find_by_mlb_id(db_session, pitcher_id_mlb)
        batter_id = cls.find_by_mlb_id(db_session, batter_id_mlb)
        return {
            "pitcher_id_bbref": pitcher_id.bbref_id,
            "pitcher_id_mlb": pitcher_id_mlb,
            "pitcher_id_db": pitcher_id.db_player_id,
            "pitcher_name": pitcher_id.mlb_name,
            "pitcher_team": at_bat_dict["pitcher_team"],
            "batter_id_bbref": batter_id.bbref_id,
            "batter_id_mlb": batter_id_mlb,
            "batter_id_db": batter_id.db_player_id,
            "batter_name": batter_id.mlb_name,
            "batter_team": at_bat_dict["batter_team"],
        }

    @classmethod
    def get_player_id_map(cls, db_session):
        all_players = db_session.query(cls).all()
        return {p.mlb_id: p.db_player_id for p in all_players}



"""Update bbref_player_id_map.json file."""
import subprocess
from collections import defaultdict

from getch import pause
from halo import Halo

from vigorish.cli.components import (
    get_random_cli_color,
    get_random_dots_spinner,
    print_heading,
    print_message,
    user_options_prompt,
    yes_no_prompt,
)
from vigorish.cli.menu_item import MenuItem
from vigorish.constants import DATA_SET_TO_NAME_MAP, EMOJIS, MENU_NUMBERS
from vigorish.tasks import ImportScrapedDataTask
from vigorish.util.result import Result

IMPORT_DATA_MESSAGE = (
    "This task uses the value of the JSON_LOCAL_FOLDER_PATH config setting to locate scraped "
    "data in the local file system from all data sets and all MLB seasons. After the search is "
    "complete, the scraped data is used to update the database."
)
IMPORT_DATA_PROMPT = "\nSelect YES to run this task\nSelect NO to return to the previous menu"
OVERWRITE_DATA_MESSAGE = (
    'By default, data that is already considered "scraped" within the database will not be '
    "updated with the data found in the local file system."
)
OVERWRITE_DATA_PROMPT = (
    "Select KEEP EXISTING DATA to use the default behavior and only update data which is not "
    'already marked as "scraped" in the database\n'
    "Select OVERWRITE EXISTING DATA to update all data in the database with the values found in "
    "the local file system"
)


class ImportScrapedData(MenuItem):
    def __init__(self, app):
        super().__init__(app)
        self.import_scraped_data = ImportScrapedDataTask(self.app)
        self.spinners = {}
        self.menu_item_text = "Import Scraped Data from Local Folders"
        self.menu_item_emoji = EMOJIS.get("HONEY_POT")

    def launch(self):
        if not self.prompt_user_import_data():
            return Result.Ok(True)
        self.subscribe_to_events()
        result = self.prompt_user_overwrite_data()
        if result.failure:
            return Result.Ok(True)
        overwrite_existing_data = result.value == "OVERWRITE"
        result = self.import_scraped_data.execute(overwrite_existing_data)
        self.unsubscribe_from_events()
        if result.failure:
            self.display_error_messages(result.error)
        return Result.Ok(True)

    def update_menu_heading(self, current_action):
        new_heading = (
            f"Import Scraped Data: {current_action}"
            if self.app.db_setup_complete
            else f"Setup Database: {current_action}"
        )
        subprocess.run(["clear"])
        print_heading(new_heading, fg="bright_yellow")

    def prompt_user_import_data(self):
        self.update_menu_heading("Run Task?")
        print_message(IMPORT_DATA_MESSAGE, fg="bright_yellow", bold=True)
        return yes_no_prompt(IMPORT_DATA_PROMPT, wrap=False)

    def prompt_user_overwrite_data(self):
        self.update_menu_heading("Overwrite Existing Data?")
        print_message(OVERWRITE_DATA_MESSAGE, fg="bright_yellow", bold=True)
        choices = {
            f"{MENU_NUMBERS.get(1)}  KEEP EXISTING DATA": "KEEP",
            f"{MENU_NUMBERS.get(2)}  OVERWRITE EXISTING DATA": "OVERWRITE",
            f"{EMOJIS.get('BACK')} Return to Admin/Tasks Menu": None,
        }
        return user_options_prompt(choices, OVERWRITE_DATA_PROMPT, clear_screen=False)

    def display_error_messages(self, error_message):
        self.update_menu_heading("Error!")
        print_message(error_message)
        pause(message="\nPress any key to continue...")

    def error_occurred(self, error_message, data_set, year):
        self.update_menu_heading("Error!")
        error_message = f"Error occurred while updating {DATA_SET_TO_NAME_MAP[data_set]} for MLB {year}"
        self.spinners[year][data_set].fail(error_message)

    def search_local_files_start(self):
        self.update_menu_heading("In Progress...")
        self.spinners["default"] = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
        self.spinners["default"].text = "Searching local folder for scraped data..."
        self.spinners["default"].start()

    def import_scraped_data_start(self):
        self.spinners["default"].stop()
        self.update_menu_heading("In Progress...")

    def import_scraped_data_for_year_start(self, year):
        self.update_menu_heading("In Progress...")
        self.spinners[year] = defaultdict(lambda: Halo())

    def import_scraped_data_set_start(self, data_set, year):
        spinner = self.spinners[year][data_set]
        spinner.spinner = get_random_dots_spinner()
        spinner.color = get_random_cli_color()
        spinner.text = f"Updating {DATA_SET_TO_NAME_MAP[data_set]} for MLB {year}..."
        spinner.start()

    def import_scraped_data_set_complete(self, data_set, year):
        import_complete = f"Successfully updated {DATA_SET_TO_NAME_MAP[data_set]} for MLB {year}!"
        self.spinners[year][data_set].succeed(import_complete)

    def import_combined_game_data_start(self, year):
        self.spinners["combined_data"] = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
        self.spinners["combined_data"].text = f"Updating combined game data for MLB {year}..."
        self.spinners["combined_data"].start()

    def import_combined_game_data_complete(self, year):
        import_complete = f"Successfully updated combined game data for MLB {year}!"
        if "combined_data" not in self.spinners:
            self.spinners["combined_data"] = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
            self.spinners["combined_data"].text = f"Updating combined game data for MLB {year}..."
            self.spinners["combined_data"].start()
        self.spinners["combined_data"].succeed(import_complete)

    def import_scraped_data_complete(self):
        self.update_menu_heading("Complete!")
        print_message("Successfully imported all scraped data from local files", fg="bright_yellow", bold=True)
        pause(message="\nPress any key to continue...")

    def subscribe_to_events(self):
        self.import_scraped_data.events.error_occurred += self.error_occurred
        self.import_scraped_data.events.search_local_files_start += self.search_local_files_start
        self.import_scraped_data.events.import_scraped_data_start += self.import_scraped_data_start
        self.import_scraped_data.events.import_scraped_data_complete += self.import_scraped_data_complete
        self.import_scraped_data.events.import_scraped_data_for_year_start += self.import_scraped_data_for_year_start
        self.import_scraped_data.events.import_scraped_data_set_start += self.import_scraped_data_set_start
        self.import_scraped_data.events.import_scraped_data_set_complete += self.import_scraped_data_set_complete
        self.import_scraped_data.events.import_combined_game_data_start += self.import_combined_game_data_start
        self.import_scraped_data.events.import_combined_game_data_complete += self.import_combined_game_data_complete

    def unsubscribe_from_events(self):
        self.import_scraped_data.events.error_occurred -= self.error_occurred
        self.import_scraped_data.events.search_local_files_start -= self.search_local_files_start
        self.import_scraped_data.events.import_scraped_data_start -= self.import_scraped_data_start
        self.import_scraped_data.events.import_scraped_data_complete -= self.import_scraped_data_complete
        self.import_scraped_data.events.import_scraped_data_for_year_start -= self.import_scraped_data_for_year_start
        self.import_scraped_data.events.import_scraped_data_set_start -= self.import_scraped_data_set_start
        self.import_scraped_data.events.import_scraped_data_set_complete -= self.import_scraped_data_set_complete
        self.import_scraped_data.events.import_combined_game_data_start -= self.import_combined_game_data_start
        self.import_scraped_data.events.import_combined_game_data_complete -= self.import_combined_game_data_complete



from pathlib import Path

from tabulate import tabulate

from vigorish.cli.components.viewers.display_table import DisplayTable
from vigorish.cli.components.viewers.table_viewer import TableViewer
from vigorish.util.list_helpers import make_chunked_list


class CsvFileTableViewer(TableViewer):
    def __init__(
        self,
        csv_filepath,
        prompt,
        confirm_only=False,
        yes_choice="YES",
        no_choice="NO",
        rows_per_page=10,
        heading=None,
        heading_color=None,
        message_color=None,
        table_color=None,
    ):
        table_list = self.create_tables(csv_filepath, rows_per_page, heading)
        super().__init__(
            table_list=table_list,
            prompt=prompt,
            confirm_only=confirm_only,
            yes_choice=yes_choice,
            no_choice=no_choice,
            heading_color=heading_color,
            message_color=message_color,
            table_color=table_color,
        )

    def create_tables(self, csv_filepath, rows_per_page, heading):
        (col_names, table_rows) = self.get_table_data_from_csv(csv_filepath)
        table_chunks = make_chunked_list(table_rows, chunk_size=rows_per_page)
        return [DisplayTable(tabulate(rows, headers=col_names), heading) for rows in table_chunks]

    def get_table_data_from_csv(self, csv_filepath):
        if isinstance(csv_filepath, str):
            csv_filepath = Path(csv_filepath)
        csv_text = csv_filepath.read_text()
        csv_rows = csv_text.split("\n")
        col_names = [col.strip() for col in csv_rows.pop(0).split(",")]
        table_rows = [row.split(",") for row in csv_rows]
        for num, row in enumerate(table_rows, start=1):
            row.insert(0, num)
        col_names.insert(0, "#")
        return (col_names, table_rows)



"""ABC for menu items that execute a command."""
import subprocess
from abc import ABC, abstractmethod

from bullet import colors

from vigorish.cli.components.util import print_heading


class MenuItem(ABC):
    # TODO: Update all menu items to use the menu_heading value

    _menu_item_text = ""
    menu_item_emoji = ""
    menu_heading = "Menu"
    background_color = colors.background["default"]
    background_on_switch = colors.background["default"]
    word_color = colors.foreground["default"]
    word_on_switch = colors.bright(colors.foreground["cyan"])
    exit_menu = False

    def __init__(self, app):
        self.app = app
        self.dotenv = app.dotenv
        self.config = app.config
        self.db_engine = app.db_engine
        self.db_session = app.db_session
        self.scraped_data = app.scraped_data

    @property
    def menu_item_text(self):
        return f"{self.menu_item_emoji} {self._menu_item_text}"

    @menu_item_text.setter
    def menu_item_text(self, menu_item_text):
        self._menu_item_text = menu_item_text

    def get_menu_heading(self, status):
        return f"{self.menu_heading}: {status}"

    def update_menu_heading(self, status, heading_color="bright_yellow"):
        subprocess.run(["clear"])
        print_heading(self.get_menu_heading(status), fg=heading_color)

    @abstractmethod
    def launch(self):
        pass



import subprocess

from getch import pause
from halo import Halo
from tabulate import tabulate

import vigorish.database as db
from vigorish.cli.components.util import (
    get_random_cli_color,
    get_random_dots_spinner,
    print_heading,
    print_message,
)
from vigorish.cli.components.viewers import DisplayPage, PageViewer
from vigorish.cli.menu_item import MenuItem
from vigorish.constants import EMOJIS
from vigorish.tasks import CalculateAvgPitchTimesTask
from vigorish.util.result import Result

STAT_NAMES = ["avg", "min", "max"]


class CalculatePitchTimes(MenuItem):
    def __init__(self, app):
        super().__init__(app)
        self.calc_pitch_times = CalculateAvgPitchTimesTask(app)
        self.menu_item_text = "Calculate Avg. Time Between Pitches"
        self.menu_item_emoji = EMOJIS.get("CLOCK")
        self.exit_menu = False
        self.spinner = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
        self.game_ids = []

    @property
    def total_games(self):
        return len(self.game_ids)

    def launch(self):
        if not self.prompt_user_run_task():
            return Result.Ok(True)
        self.subscribe_to_events()
        result = self.calc_pitch_times.execute()
        self.spinner.stop()
        self.unsubscribe_from_events()
        if result.failure:
            return result
        subprocess.run(["clear"])
        results = result.value
        time_between_pitches = db.TimeBetweenPitches.from_calc_results(self.db_session, results)
        self.display_pitch_metrics(time_between_pitches.as_dict())
        pause(message="\nPress any key to continue...")
        return Result.Ok()

    def prompt_user_run_task(self):
        subprocess.run(["clear"])
        prompt = "Would you like to run this task and calculate these metrics?"
        pages = [DisplayPage(page, None) for page in self.get_task_description_pages()]
        page_viewer = PageViewer(pages, prompt=prompt, text_color="bright_yellow")
        return page_viewer.launch()

    def find_eligible_games_start(self):
        subprocess.run(["clear"])
        self.spinner.text = "Identifying games which meet all requirements for task..."
        self.spinner.start()

    def find_eligible_games_complete(self, game_ids):
        self.spinner.stop()
        self.game_ids = game_ids
        self.spinner = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
        self.spinner.text = self.get_progress_text(0)
        self.spinner.start()

    def calculate_pitch_metrics_progress(self, games_complete):
        self.spinner.text = self.get_progress_text(games_complete)

    def get_progress_text(self, games_complete):
        game_id = self.game_ids[games_complete - 1]
        percent = games_complete / float(self.total_games)
        return f"Calculating time between pitches... (Game ID: {game_id}) {percent:.0%}..."

    def display_pitch_metrics(self, metrics):
        heading = f"Avg. Time Between Pitches for {len(self.game_ids)} MLB Games"
        table = self.construct_stats_table(metrics)
        print_heading(heading, fg="bright_yellow")
        print_message(table, wrap=False, fg="bright_cyan")

    def construct_stats_table(self, metrics):
        table_data = []
        for stat_name in STAT_NAMES:
            row = {"stat_name": stat_name}
            for group_name, pitch_metrics in metrics.items():
                if group_name == "timestamp":
                    continue
                row[group_name] = pitch_metrics[stat_name]
            table_data.append(row)
        return tabulate(table_data, headers="keys", numalign="right", stralign="right")

    def subscribe_to_events(self):
        self.calc_pitch_times.events.find_eligible_games_start += self.find_eligible_games_start
        self.calc_pitch_times.events.find_eligible_games_complete += self.find_eligible_games_complete
        self.calc_pitch_times.events.calculate_pitch_metrics_progress += self.calculate_pitch_metrics_progress

    def unsubscribe_from_events(self):
        self.calc_pitch_times.events.find_eligible_games_start -= self.find_eligible_games_start
        self.calc_pitch_times.events.find_eligible_games_complete -= self.find_eligible_games_complete
        self.calc_pitch_times.events.calculate_pitch_metrics_progress -= self.calculate_pitch_metrics_progress

    def get_task_description_pages(self):
        return [
            [
                (
                    "This task first identifies games where all data has been scraped and the "
                    "process of reconciling and combining the scraped data was entirely "
                    "successful with zero data errors of any type (e.g., zero missing/extra "
                    "pitches, no invalid PitchFX data).\n"
                ),
                (
                    "Next, each game is stepped through pitch-by-pitch and the length of time "
                    "between each pitch is calculated and stored using the timestamps "
                    "available on PitchFX data."
                ),
            ],
            [
                "There are three distinct types of between-pitch durations which are collected:\n",
                "1. (between pitches)",
                "time between pitches in the same at bat\n",
                "2. (between at bats)",
                (
                    "time between the last pitch of an at bat and the first pitch of the next at "
                    "bat in the same inning"
                ),
            ],
            [
                "3. (between innings)",
                (
                    "time between the last pitch of the last at bat in an inning and the first "
                    "pitch of the next at bat in the subsequent inning\n"
                ),
                (
                    "Finally, the data sets are processed to remove outliers before calculating "
                    "the average, maximum and minimum durations for each category."
                ),
            ],
        ]



"""Menu that allows the user to view and modify all settings in vig.config.json."""
import subprocess

from vigorish.cli.components import print_heading, print_message, yes_no_prompt
from vigorish.cli.menu_item import MenuItem
from vigorish.cli.menu_items.change_config_setting import ChangeConfigSettting
from vigorish.constants import EMOJIS
from vigorish.util.result import Result


class ConfigSetting(MenuItem):
    def __init__(self, app, setting_name):
        super().__init__(app)
        self.setting_name = setting_name
        self.setting = self.config.all_settings.get(setting_name)
        self.setting_name_title = self.setting.setting_name_title
        self.current_settings = self.setting.current_settings_report
        self.menu_item_text = self.setting.setting_name_title
        self.menu_item_emoji = EMOJIS.get("GEAR")
        self.data_type = self.setting.data_type

    def launch(self):
        subprocess.run(["clear"])
        setting_heading = f"Setting: {self.setting_name_title} (Type: {self.data_type.name})"
        print_heading(setting_heading, fg="bright_magenta")
        self.print_description()
        print_message(self.current_settings, wrap=False, fg="bright_yellow", bold=True)
        if yes_no_prompt("\nChange current setting?"):
            change_setting_menu = ChangeConfigSettting(self.app, self.setting_name)
            return change_setting_menu.launch()
        return Result.Ok(self.exit_menu)

    def print_description(self):
        if isinstance(self.setting.description, list):
            for line in self.setting.description:
                print_message(line)
        else:
            print_message(self.setting.description)
        print()



import subprocess
from datetime import datetime

from bullet import colors, Input
from dateutil import parser as date_parser
from getch import pause

from vigorish.cli.components.util import print_heading, print_message

ERROR_HELP_MESSAGE = (
    "You can use any format recognized by dateutil.parser. For example, all of the strings below "
    "are valid ways to represent the same date:\n"
)
EXAMPLES = '"2018-5-13" -or- "05/13/2018" -or- "May 13 2018"\n'


class DateInput(Input):
    """Prompt user for a date value until successfully parsed.

    Date can be provided in any format recognized by dateutil.parser.

    Args:
        prompt (str): Required. Text to display to user before input prompt.
        heading (str): Optional. Bold, underlined text to display above input prompt.
        heading_color (str): Optional. The color of the heading.
        default (datetime): Optional. Default value to use if user provides no input.
        format_str (str): Format string used to display default value, defaults to '%m/%d/%Y'
        indent (int): Distance between left-boundary and start of prompt.
        word_color (str): Optional. The color of the prompt and user input.

    Returns:
        Date: A new Date prompt.
    """

    def __init__(
        self,
        prompt: str,
        heading: str = None,
        heading_color: str = colors.foreground["default"],
        default: datetime = None,
        format_str: str = "%m/%d/%Y",
        indent: int = 0,
        word_color: str = colors.foreground["default"],
    ):
        self.heading = heading
        self.heading_color = heading_color
        if default:
            default = default.strftime(format_str)
        super().__init__(prompt, default=default, indent=indent, word_color=word_color)

    def launch(self):
        while True:
            subprocess.run(["clear"])
            if self.heading:
                print_heading(self.heading, fg=self.heading_color)
            result = super().launch()
            if not result:
                continue
            try:
                return date_parser.parse(result)
            except ValueError:
                error = f'\nError: "{result}" could not be parsed as a valid date.'
                print_heading(error, fg="bright_red")
                print_message(ERROR_HELP_MESSAGE, fg="bright_red")
                print_message(EXAMPLES, fg="bright_red")
                pause(message="Press any key to continue...")



import os
from datetime import datetime, timezone
from pathlib import Path
from zipfile import ZIP_DEFLATED, ZipFile

from events import Events

from vigorish.models.bat_stats import BatStats, BatStatsCsvRow
from vigorish.models.pitch_stats import PitchStats, PitchStatsCsvRow
from vigorish.models.pitchfx import PitchFx, PitchFxCsvRow
from vigorish.models.status_date import DateScrapeStatus, DateScrapeStatusCsvRow
from vigorish.models.status_game import GameScrapeStatus, GameScrapeStatusCsvRow
from vigorish.models.status_pitch_appearance import PitchAppScrapeStatus, PitchAppScrapeStatusCsvRow
from vigorish.tasks.base import Task
from vigorish.util.dataclass_helpers import serialize_db_object_to_csv
from vigorish.util.dt_format_strings import CSV_UTC, DATE_ONLY, DT_AWARE, FILE_TIMESTAMP
from vigorish.util.numeric_helpers import ONE_PERCENT
from vigorish.util.result import Result

DB_MODEL_TO_CSV_MAP = {
    DateScrapeStatus: {"dataclass": DateScrapeStatusCsvRow, "date_format": DATE_ONLY},
    GameScrapeStatus: {"dataclass": GameScrapeStatusCsvRow, "date_format": DATE_ONLY},
    PitchAppScrapeStatus: {"dataclass": PitchAppScrapeStatusCsvRow, "date_format": DT_AWARE},
    BatStats: {"dataclass": BatStatsCsvRow, "date_format": None},
    PitchStats: {"dataclass": PitchStatsCsvRow, "date_format": None},
    PitchFx: {"dataclass": PitchFxCsvRow, "date_format": CSV_UTC},
}


class BackupDatabaseTask(Task):
    def __init__(self, app):
        super().__init__(app)
        self.csv_folder = None
        self.events = Events(
            (
                "backup_database_start",
                "backup_table_start",
                "backup_table_progress",
                "backup_table_complete",
                "create_zip_file_start",
                "create_zip_file_complete",
                "backup_database_complete",
            )
        )

    def execute(self):
        csv_map = self.get_csv_map()
        self.events.backup_database_start(len(csv_map))
        for table_num, (table, csv_file) in enumerate(csv_map.items(), start=1):
            self.events.backup_table_start(table_num, table)
            self.export_table_to_csv(table, csv_file)
            self.events.backup_table_complete(table_num, table)
        self.events.create_zip_file_start()
        zip_file = self.create_zip_file(csv_map)
        self.remove_csv_files(csv_map)
        self.events.create_zip_file_complete()
        self.events.backup_database_complete()
        return Result.Ok(zip_file)

    def get_csv_map(self):
        self.csv_folder = self.create_csv_folder()
        return {
            DateScrapeStatus: self.csv_folder.joinpath("scrape_status_date.csv"),
            GameScrapeStatus: self.csv_folder.joinpath("scrape_status_game.csv"),
            PitchAppScrapeStatus: self.csv_folder.joinpath("scrape_status_pitch_app.csv"),
            BatStats: self.csv_folder.joinpath("bat_stats.csv"),
            PitchStats: self.csv_folder.joinpath("pitch_stats.csv"),
            PitchFx: self.csv_folder.joinpath("pitchfx.csv"),
        }

    def create_csv_folder(self):
        backup_folder = self.app.get_current_setting("DB_BACKUP_FOLDER_PATH")
        folder_name = f"{datetime.now(timezone.utc).strftime(FILE_TIMESTAMP)}"
        if os.environ.get("ENV") == "TEST":
            folder_name = "__timestamp__"
        csv_folder = Path(backup_folder).joinpath(folder_name)
        csv_folder.mkdir(parents=True)
        return csv_folder

    def export_table_to_csv(self, table, csv_file):
        self.append_to_csv(csv_file, text=self.get_csv_column_names(table))
        total_rows = self.app.get_total_number_of_rows(table)
        chunk_size, chunk_count, row_count, last_reported = 10000, 0, 0, 0
        while True:
            start = chunk_size * chunk_count
            stop = min(chunk_size * (chunk_count + 1), total_rows)
            table_chunk = self.db_session.query(table).slice(start, stop).all()
            if not table_chunk:
                break
            self.append_to_csv(csv_file, text=self.create_csv_rows(table_chunk, table))
            row_count += len(table_chunk)
            chunk_count += 1
            last_reported = self.report_progress(row_count, total_rows, last_reported)
            if row_count == total_rows:
                break

    def append_to_csv(self, csv_file, text):
        with csv_file.open("a") as csv:
            csv.write(f"{text}\n")

    def get_csv_column_names(self, table):
        csv_dataclass = DB_MODEL_TO_CSV_MAP[table]["dataclass"]
        return ",".join(list(csv_dataclass.__dataclass_fields__.keys()))

    def create_csv_rows(self, table_chunk, table):
        csv_rows = [
            serialize_db_object_to_csv(
                db_obj,
                dataclass=DB_MODEL_TO_CSV_MAP[table]["dataclass"],
                date_format=DB_MODEL_TO_CSV_MAP[table]["date_format"],
            )
            for db_obj in table_chunk
        ]
        return "\n".join(csv_rows)

    def report_progress(self, count, total, last_reported):
        percent_complete = count / float(total)
        if percent_complete - last_reported > ONE_PERCENT:
            self.events.backup_table_progress(percent_complete)
            last_reported = percent_complete
        return last_reported

    def create_zip_file(self, csv_map):
        zip_filename = f"{self.csv_folder.name}.zip"
        zip_file = self.csv_folder.parent.joinpath(zip_filename)
        with ZipFile(zip_file, "w", ZIP_DEFLATED) as zip:
            for csv_file in csv_map.values():
                arcname = f"{self.csv_folder.name}/{csv_file.name}"
                zip.write(csv_file, arcname=arcname)
        return zip_file

    def remove_csv_files(self, csv_map):
        for csv_file in csv_map.values():
            csv_file.unlink()
        self.csv_folder.rmdir()



"""Shared functions and menus for CLI."""
import subprocess
from functools import partial
from random import randint

import click
from bullet import Bullet, colors, ScrollBar
from getch import pause

import vigorish.database as db
from vigorish.constants import CLI_COLORS, FIGLET_FONTS
from vigorish.util.result import Result
from vigorish.util.string_helpers import wrap_text


def create_bullet_prompt(prompt, choices, wrap=True, max_line_len=70):
    if wrap:
        prompt = wrap_text(prompt, max_line_len)
    return Bullet(
        prompt,
        choices,
        bullet="",
        shift=1,
        indent=0,
        margin=2,
        bullet_color=colors.foreground["default"],
        background_color=colors.foreground["default"],
        background_on_switch=colors.foreground["default"],
        word_color=colors.foreground["default"],
        word_on_switch=colors.bright(colors.foreground["cyan"]),
    )


def create_scrolling_prompt(prompt, choices, wrap=True, max_line_len=70):
    if wrap:
        prompt = wrap_text(prompt, max_line_len)
    return ScrollBar(
        prompt,
        choices,
        height=8,
        pointer="",
        shift=1,
        indent=0,
        margin=2,
        pointer_color=colors.foreground["default"],
        word_color=colors.foreground["default"],
        word_on_switch=colors.bright(colors.foreground["cyan"]),
        background_color=colors.background["default"],
        background_on_switch=colors.background["default"],
    )


def get_random_cli_color():
    colors = ["red", "blue", "green", "cyan", "magenta", "yellow"]
    return colors[randint(0, len(colors) - 1)]


def get_random_bright_cli_color():
    colors = [
        "bright_red",
        "bright_blue",
        "bright_green",
        "bright_cyan",
        "bright_magenta",
        "bright_yellow",
    ]
    return colors[randint(0, len(colors) - 1)]


def get_random_dots_spinner():
    return f"dots{randint(2, 9)}"


def get_random_figlet_font():
    return FIGLET_FONTS[randint(0, len(FIGLET_FONTS) - 1)]


def print_message(message, wrap=True, max_line_len=70, fg=None, bg=None, bold=None, underline=None):
    if (fg and fg not in CLI_COLORS) or (bg and bg not in CLI_COLORS):
        fg = None
        bg = None
    if wrap:
        message = wrap_text(message, max_len=max_line_len)
    click.secho(message, fg=fg, bg=bg, bold=bold, underline=underline)


def print_heading(message, fg=None, bg=None):
    print_message(f"{message}\n", wrap=False, fg=fg, bg=bg, bold=True, underline=True)


print_error = partial(print_message, fg="bright_red", bg=None, bold=True, underline=None)
print_success = partial(print_message, fg="bright_green", bg=None, bold=True, underline=None)


def validate_scrape_dates(db_session, start_date, end_date):
    result = db.Season.validate_date_range(db_session, start_date, end_date)
    if result.success:
        season = result.value
        return Result.Ok(season)
    print_heading("\nError! Invalid value for start and/or end dates", fg="bright_red")
    for s in result.error:
        print_message(s, fg="bright_red")
    pause(message="\nPress any key to continue...")
    return Result.Fail("")


def shutdown_cli_immediately():
    subprocess.run(["clear"])
    print_heading("Restart Required!", fg="bright_magenta")
    warning = "Application must be restarted for these changes to take effect."
    print_message(warning, fg="bright_magenta", bold=True)
    print_message("Shutting down vigorish!\n", fg="bright_magenta", bold=True)
    pause(message="Press any key to continue...")
    subprocess.run(["clear"])
    exit(0)



"""Db model that describes a MLB season and tracks data scraping progress."""
from datetime import date

from sqlalchemy import Column, DateTime, Integer
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship
from sqlalchemy.types import Enum

import vigorish.database as db
from vigorish.enums import SeasonType
from vigorish.util.datetime_util import get_date_range as get_date_range_util
from vigorish.util.dt_format_strings import DATE_ONLY
from vigorish.util.exceptions import InvalidSeasonException
from vigorish.util.result import Result


class Season(db.Base):
    """Database model that describes a MLB season and tracks data scraping progress."""

    __tablename__ = "season"
    id = Column(Integer, primary_key=True)
    year = Column(Integer, nullable=False)
    start_date = Column(DateTime, nullable=False)
    end_date = Column(DateTime, nullable=False)
    asg_date = Column(DateTime, default=date.min)
    season_type = Column(Enum(SeasonType), default=SeasonType.NONE)

    dates = relationship("DateScrapeStatus", backref="season")
    games = relationship("GameScrapeStatus", backref="season")
    pitch_apps = relationship("PitchAppScrapeStatus", backref="season")
    scrape_jobs = relationship("ScrapeJob", backref="season")
    pitch_app_status = relationship(
        "Season_PitchApp_View",
        backref="original",
        uselist=False,
        primaryjoin="Season.id==Season_PitchApp_View.id",
        foreign_keys="Season_PitchApp_View.id",
    )
    game_status = relationship(
        "Season_Game_View",
        backref="original",
        uselist=False,
        primaryjoin="Season.id==Season_Game_View.id",
        foreign_keys="Season_Game_View.id",
    )
    date_status = relationship(
        "Season_Date_View",
        backref="original",
        uselist=False,
        primaryjoin="Season.id==Season_Date_View.id",
        foreign_keys="Season_Date_View.id",
    )

    @hybrid_property
    def name(self):
        return f'MLB {self.year} {self.season_type.replace("_", " ").title()}'

    @hybrid_property
    def start_date_str(self):
        return self.start_date.strftime(DATE_ONLY)

    @hybrid_property
    def end_date_str(self):
        return self.end_date.strftime(DATE_ONLY)

    @hybrid_property
    def asg_date_str(self):
        return self.asg_date.strftime(DATE_ONLY)

    @hybrid_property
    def total_days(self):
        today = date.today()
        if today.year == self.year and today < self.end_date.date():
            return (today - self.start_date.date()).days
        return self.date_status.total_days

    @hybrid_property
    def total_days_scraped_bbref(self):
        return (
            self.date_status.total_scraped_daily_dash_bbref
            if self.date_status and self.date_status.total_scraped_daily_dash_bbref
            else 0
        )

    @hybrid_property
    def percent_complete_bbref_games_for_date(self):
        return self.total_days_scraped_bbref / float(self.total_days) if self.total_days > 0 else 0.0

    @hybrid_property
    def scraped_all_bbref_games_for_date(self):
        return self.total_days == self.total_days_scraped_bbref if self.total_days > 0 else False

    @hybrid_property
    def total_days_scraped_brooks(self):
        return (
            self.date_status.total_scraped_daily_dash_brooks
            if self.date_status and self.date_status.total_scraped_daily_dash_brooks
            else 0
        )

    @hybrid_property
    def percent_complete_brooks_games_for_date(self):
        return self.total_days_scraped_brooks / float(self.total_days) if self.total_days > 0 else 0.0

    @hybrid_property
    def scraped_all_brooks_games_for_date(self):
        return self.total_days == self.total_days_scraped_brooks if self.total_days > 0 else False

    @hybrid_property
    def total_games(self):
        return self.game_status.total_games if (self.game_status and self.game_status.total_games) else 0

    @hybrid_property
    def total_games_combined_success(self):
        return (
            self.game_status.total_combined_data_success
            if self.game_status and self.game_status.total_combined_data_success
            else 0
        )

    @hybrid_property
    def total_games_combined_fail(self):
        return (
            self.game_status.total_combined_data_fail
            if self.game_status and self.game_status.total_combined_data_fail
            else 0
        )

    @hybrid_property
    def total_games_combined(self):
        return self.total_games_combined_success + self.total_games_combined_fail

    @hybrid_property
    def total_bbref_boxscores_scraped(self):
        return (
            self.game_status.total_scraped_bbref_boxscore
            if self.game_status and self.game_status.total_scraped_bbref_boxscore
            else 0
        )

    @hybrid_property
    def percent_complete_bbref_boxscores_scraped(self):
        return self.total_bbref_boxscores_scraped / float(self.total_games) if self.total_games > 0 else 0.0

    @hybrid_property
    def scraped_all_bbref_boxscores(self):
        return self.total_games == self.total_bbref_boxscores_scraped if self.total_games > 0 else False

    @hybrid_property
    def total_brooks_pitch_logs_scraped(self):
        return (
            self.game_status.total_scraped_brooks_pitch_logs
            if self.game_status and self.game_status.total_scraped_brooks_pitch_logs
            else 0
        )

    @hybrid_property
    def percent_complete_brooks_pitch_logs(self):
        return self.total_brooks_pitch_logs_scraped / float(self.total_games) if self.total_games > 0 else 0.0

    @hybrid_property
    def scraped_all_brooks_pitch_logs(self):
        return self.total_games == self.total_brooks_pitch_logs_scraped if self.total_games > 0 else False

    @hybrid_property
    def pitch_app_count_bbref(self):
        return (
            self.game_status.total_pitch_app_count_bbref
            if self.game_status and self.game_status.total_pitch_app_count_bbref
            else 0
        )

    @hybrid_property
    def pitch_app_count_brooks(self):
        return (
            self.game_status.total_pitch_app_count_brooks
            if self.game_status and self.game_status.total_pitch_app_count_brooks
            else 0
        )

    @hybrid_property
    def total_pitch_count_bbref(self):
        return (
            self.game_status.total_pitch_count_bbref
            if self.game_status and self.game_status.total_pitch_count_bbref
            else 0
        )

    @hybrid_property
    def pitch_app_count_pitchfx(self):
        return (
            self.pitch_app_status.total_pitchfx if self.pitch_app_status and self.pitch_app_status.total_pitchfx else 0
        )

    @hybrid_property
    def total_pitch_apps_scraped_pitchfx(self):
        return (
            self.pitch_app_status.total_pitchfx_scraped
            if self.pitch_app_status and self.pitch_app_status.total_pitchfx_scraped
            else 0
        )

    @hybrid_property
    def total_pitch_apps_no_pitchfx_data(self):
        return (
            self.pitch_app_status.total_no_pitchfx_data
            if self.pitch_app_status and self.pitch_app_status.total_no_pitchfx_data
            else 0
        )

    @hybrid_property
    def total_pitch_apps_with_pitchfx_data(self):
        return self.pitch_app_count_pitchfx - self.total_pitch_apps_no_pitchfx_data

    @hybrid_property
    def total_pitch_apps_combined_data(self):
        return (
            self.pitch_app_status.total_combined_pitchfx_bbref_data
            if self.pitch_app_status and self.pitch_app_status.total_combined_pitchfx_bbref_data
            else 0
        )

    @hybrid_property
    def total_pitch_apps_pitchfx_error(self):
        return (
            self.pitch_app_status.total_pitchfx_error
            if self.pitch_app_status and self.pitch_app_status.total_pitchfx_error
            else 0
        )

    @hybrid_property
    def total_pitch_apps_invalid_pitchfx(self):
        return (
            self.pitch_app_status.total_invalid_pitchfx
            if self.pitch_app_status and self.pitch_app_status.total_invalid_pitchfx
            else 0
        )

    @hybrid_property
    def total_pitch_apps_pitchfx_is_valid(self):
        return sum(not (pfx.pitchfx_error or pfx.invalid_pitchfx) for pfx in self.pitch_apps)

    @hybrid_property
    def total_pitch_count_pitch_logs(self):
        return (
            self.pitch_app_status.total_pitch_count_pitch_log
            if self.pitch_app_status and self.pitch_app_status.total_pitch_count_pitch_log
            else 0
        )

    @hybrid_property
    def total_pitch_count_bbref_audited(self):
        return (
            self.pitch_app_status.total_pitch_count_bbref
            if self.pitch_app_status and self.pitch_app_status.total_pitch_count_bbref
            else 0
        )

    @hybrid_property
    def total_pitch_count_pitchfx(self):
        return (
            self.pitch_app_status.total_pitch_count_pitchfx
            if self.pitch_app_status and self.pitch_app_status.total_pitch_count_pitchfx
            else 0
        )

    @hybrid_property
    def total_pitch_count_pitchfx_audited(self):
        return (
            self.pitch_app_status.total_pitch_count_pitchfx_audited
            if self.pitch_app_status and self.pitch_app_status.total_pitch_count_pitchfx_audited
            else 0
        )

    @hybrid_property
    def total_missing_pitchfx_count(self):
        return (
            self.pitch_app_status.total_missing_pitchfx_count
            if self.pitch_app_status and self.pitch_app_status.total_missing_pitchfx_count
            else 0
        )

    @hybrid_property
    def total_removed_pitchfx_count(self):
        return (
            self.pitch_app_status.total_removed_pitchfx_count
            if self.pitch_app_status and self.pitch_app_status.total_removed_pitchfx_count
            else 0
        )

    @hybrid_property
    def total_batters_faced_bbref(self):
        return (
            self.pitch_app_status.total_batters_faced_bbref
            if self.pitch_app_status and self.pitch_app_status.total_batters_faced_bbref
            else 0
        )

    @hybrid_property
    def total_batters_faced_pitchfx(self):
        return (
            self.pitch_app_status.total_batters_faced_pitchfx
            if self.pitch_app_status and self.pitch_app_status.total_batters_faced_pitchfx
            else 0
        )

    @hybrid_property
    def total_at_bats_missing_pitchfx(self):
        return (
            self.pitch_app_status.total_at_bats_missing_pitchfx
            if self.pitch_app_status and self.pitch_app_status.total_at_bats_missing_pitchfx
            else 0
        )

    @hybrid_property
    def total_at_bats_removed_pitchfx(self):
        return (
            self.pitch_app_status.total_at_bats_removed_pitchfx
            if self.pitch_app_status and self.pitch_app_status.total_at_bats_removed_pitchfx
            else 0
        )

    @hybrid_property
    def total_at_bats_pitchfx_error(self):
        return (
            self.pitch_app_status.total_at_bats_pitchfx_error
            if self.pitch_app_status and self.pitch_app_status.total_at_bats_pitchfx_error
            else 0
        )

    @hybrid_property
    def total_at_bats_invalid_pitchfx(self):
        return (
            self.pitch_app_status.total_at_bats_invalid_pitchfx
            if self.pitch_app_status and self.pitch_app_status.total_at_bats_invalid_pitchfx
            else 0
        )

    @hybrid_property
    def scraped_all_pitchfx_logs(self):
        return (
            True
            if not self.pitch_apps
            else False
            if not self.scraped_all_brooks_pitch_logs
            else self.pitch_app_count_pitchfx == self.total_pitch_apps_scraped_pitchfx
        )

    @hybrid_property
    def combined_data_for_all_pitchfx_logs(self):
        return (
            False
            if not self.pitch_apps or not self.scraped_all_pitchfx_logs
            else self.pitch_app_count_pitchfx == self.total_pitch_apps_combined_data
        )

    @hybrid_property
    def pitchfx_error_for_any_pitchfx_logs(self):
        return (
            False
            if not self.pitch_apps or not self.scraped_all_pitchfx_logs
            else any((pfx.pitchfx_error or pfx.invalid_pitchfx) for pfx in self.pitch_apps)
        )

    @hybrid_property
    def pitchfx_is_valid_for_all_pitchfx_logs(self):
        return (
            True
            if not self.pitch_apps or not self.scraped_all_pitchfx_logs
            else all(not (pfx.pitchfx_error or pfx.invalid_pitchfx) for pfx in self.pitch_apps)
        )

    @hybrid_property
    def percent_complete_pitchfx_logs_scraped(self):
        return (
            self.total_pitch_apps_scraped_pitchfx / float(self.pitch_app_count_pitchfx)
            if self.pitch_app_count_pitchfx > 0
            else 0.0
        )

    def __repr__(self):
        return f"<Season name={self.name}, id={self.id})>"

    def as_dict(self):
        return {
            "name": self.name,
            "year": self.year,
            "start_date_str": self.start_date_str,
            "end_date_str": self.end_date_str,
            "asg_date_str": self.asg_date_str,
            "total_days": self.total_days,
            "total_days_scraped_bbref": self.total_days_scraped_bbref,
            "percent_complete_bbref_games_for_date": self.percent_complete_bbref_games_for_date,
            "scraped_all_bbref_games_for_date": self.scraped_all_bbref_games_for_date,
            "total_days_scraped_brooks": self.total_days_scraped_brooks,
            "percent_complete_brooks_games_for_date": self.percent_complete_brooks_games_for_date,
            "scraped_all_brooks_games_for_date": self.scraped_all_brooks_games_for_date,
            "total_games": self.total_games,
            "total_games_combined_success": self.total_games_combined_success,
            "total_games_combined_fail": self.total_games_combined_fail,
            "total_games_combined": self.total_games_combined,
            "total_bbref_boxscores_scraped": self.total_bbref_boxscores_scraped,
            "percent_complete_bbref_boxscores_scraped": (f"{self.percent_complete_bbref_boxscores_scraped:01.0f}%"),
            "scraped_all_bbref_boxscores": self.scraped_all_bbref_boxscores,
            "total_brooks_pitch_logs_scraped": self.total_brooks_pitch_logs_scraped,
            "percent_complete_brooks_pitch_logs": (f"{self.percent_complete_brooks_pitch_logs:01.0f}%"),
            "scraped_all_brooks_pitch_logs": self.scraped_all_brooks_pitch_logs,
            "pitch_app_count_bbref": self.pitch_app_count_bbref,
            "pitch_app_count_brooks": self.pitch_app_count_brooks,
            "pitch_app_count_pitchfx": self.pitch_app_count_pitchfx,
            "total_pitch_apps_scraped_pitchfx": self.total_pitch_apps_scraped_pitchfx,
            "total_pitch_apps_no_pitchfx_data": self.total_pitch_apps_no_pitchfx_data,
            "total_pitch_apps_with_pitchfx_data": self.total_pitch_apps_with_pitchfx_data,
            "total_pitch_apps_combined_data": self.total_pitch_apps_combined_data,
            "total_pitch_apps_pitchfx_error": self.total_pitch_apps_pitchfx_error,
            "total_pitch_apps_invalid_pitchfx": self.total_pitch_apps_invalid_pitchfx,
            "total_pitch_apps_pitchfx_is_valid": self.total_pitch_apps_pitchfx_is_valid,
            "total_pitch_count_pitch_logs": self.total_pitch_count_pitch_logs,
            "total_pitch_count_bbref": self.total_pitch_count_bbref,
            "total_pitch_count_bbref_audited": self.total_pitch_count_bbref_audited,
            "total_pitch_count_pitchfx": self.total_pitch_count_pitchfx,
            "total_pitch_count_pitchfx_audited": self.total_pitch_count_pitchfx_audited,
            "total_missing_pitchfx_count": self.total_missing_pitchfx_count,
            "total_removed_pitchfx_count": self.total_removed_pitchfx_count,
            "total_batters_faced_bbref": self.total_batters_faced_bbref,
            "total_batters_faced_pitchfx": self.total_batters_faced_pitchfx,
            "total_at_bats_missing_pitchfx": self.total_at_bats_missing_pitchfx,
            "total_at_bats_removed_pitchfx": self.total_at_bats_removed_pitchfx,
            "total_at_bats_pitchfx_error": self.total_at_bats_pitchfx_error,
            "total_at_bats_invalid_pitchfx": self.total_at_bats_invalid_pitchfx,
            "scraped_all_pitchfx_logs": self.scraped_all_pitchfx_logs,
            "combined_data_for_all_pitchfx_logs": self.combined_data_for_all_pitchfx_logs,
            "pitchfx_error_for_any_pitchfx_logs": self.pitchfx_error_for_any_pitchfx_logs,
            "pitchfx_is_valid_for_all_pitchfx_logs": self.pitchfx_is_valid_for_all_pitchfx_logs,
            "percent_complete_pitchfx_logs_scraped": self.percent_complete_pitchfx_logs_scraped,
        }

    def status_report(self):
        scraped_bbref_boxscores = "YES" if self.scraped_all_bbref_boxscores else "NO"
        scraped_brooks_pitch_logs = "YES" if self.scraped_all_brooks_pitch_logs else "NO"
        scraped_all_pitchfx_logs = "YES" if self.scraped_all_pitchfx_logs else "NO"
        combined_data_for_all_pitchfx_logs = "YES" if self.combined_data_for_all_pitchfx_logs else "NO"
        pitchfx_error_for_any_pitchfx_logs = "YES" if self.pitchfx_error_for_any_pitchfx_logs else "NO"
        return [
            (
                f"BBref Daily Dash Scraped.....................: "
                f"{self.total_days_scraped_bbref:,}/{self.total_days:,} days "
                f"({self.percent_complete_bbref_games_for_date:.0%})"
            ),
            (
                f"Brooks Daily Dash Scraped....................: "
                f"{self.total_days_scraped_brooks:,}/{self.total_days:,} days "
                f"({self.percent_complete_brooks_games_for_date:.0%})"
            ),
            (
                f"BBref Boxscores Scraped......................: {scraped_bbref_boxscores} "
                f"{self.total_bbref_boxscores_scraped:,}/{self.total_games:,}"
            ),
            (
                f"Brooks Games Scraped.........................: {scraped_brooks_pitch_logs} "
                f"{self.total_brooks_pitch_logs_scraped:,}/{self.total_games:,}"
            ),
            (
                f"PitchFx Logs Scraped.........................: {scraped_all_pitchfx_logs} "
                f"{self.total_pitch_apps_scraped_pitchfx:,}/{self.pitch_app_count_pitchfx:,} "
                f"({self.percent_complete_pitchfx_logs_scraped:.0%})"
            ),
            (
                f"Combined BBRef/PitchFX Data (Success/Total)..: "
                f"{combined_data_for_all_pitchfx_logs} "
                f"{self.total_games_combined_success:,}/{self.total_games_combined:,}"
            ),
            (
                f"Pitch App Count (BBRef/Brooks)...............: "
                f"{self.pitch_app_count_bbref:,}/{self.pitch_app_count_brooks:,}"
            ),
            (
                f"Pitch App Count (PFx/data/no data)...........: "
                f"{self.pitch_app_count_pitchfx:,}/"
                f"{self.total_pitch_apps_with_pitchfx_data:,}/"
                f"{self.total_pitch_apps_no_pitchfx_data:,}"
            ),
            (
                f"PitchFX Data Errors (Valid AB/Invalid AB)....: "
                f"{pitchfx_error_for_any_pitchfx_logs} "
                f"{self.total_pitch_apps_pitchfx_error}/"
                f"{self.total_pitch_apps_invalid_pitchfx:,}"
            ),
            (
                f"Pitch Count (BBRef/Brooks/PFx)...............: "
                f"{self.total_pitch_count_bbref:,}/"
                f"{self.total_pitch_count_pitch_logs:,}/"
                f"{self.total_pitch_count_pitchfx:,}"
            ),
            (
                "Pitch Count Audited (BBRef/PFx/Removed)......: "
                f"{self.total_pitch_count_bbref_audited:,}/"
                f"{self.total_pitch_count_pitchfx_audited:,}/"
                f"{self.total_removed_pitchfx_count:,}"
            ),
        ]

    def get_date_range(self):
        return get_date_range_util(self.start_date, self.end_date)

    def get_all_bbref_game_ids_combined_data_fail(self):
        return [game.bbref_game_id for game in self.games if game.combined_data_fail]

    @classmethod
    def find_by_year(cls, db_session, year, season_type=SeasonType.REGULAR_SEASON):
        season = db_session.query(cls).filter_by(season_type=season_type).filter_by(year=year).first()
        if not season:
            raise InvalidSeasonException(year)
        return season

    @classmethod
    def is_date_in_season(cls, db_session, check_date, season_type=SeasonType.REGULAR_SEASON):
        season = cls.find_by_year(db_session, check_date.year)
        if not season:
            error = (
                f"Database does not contain info for the MLB {check_date.year} "
                f'{season_type.replace("_", " ").title()}'
            )
            return Result.Fail(error)
        date_str = check_date.strftime(DATE_ONLY)
        if check_date < season.start_date or check_date > season.end_date:
            error = f"{date_str} is not within the scope of the {season.name}"
            return Result.Fail(error)
        return Result.Ok(season)

    @classmethod
    def validate_date_range(cls, db_session, start, end):
        if start.year != end.year:
            error = [
                "Start and end dates must both be in the same year and within "
                "the scope of that year's MLB Regular Season."
            ]
            return Result.Fail(error)
        if start > end:
            start_str = start.strftime(DATE_ONLY)
            end_str = end.strftime(DATE_ONLY)
            error = [
                '"start" must be a date before (or the same date as) "end":',
                f"start..: {start_str}",
                f"end....: {end_str}",
            ]
            return Result.Fail(error)
        season = cls.find_by_year(db_session, start.year)
        start_date_valid = cls.is_date_in_season(db_session, start).success
        end_date_valid = cls.is_date_in_season(db_session, end).success
        if not start_date_valid or not end_date_valid:
            error = [
                f"Start and end date must both be within the {season.name}:",
                f"{season.name} Start Date..: {season.start_date_str}",
                f"{season.name} End Date....: {season.end_date_str}",
            ]
            return Result.Fail(error)
        return Result.Ok(season)

    @classmethod
    def get_all_regular_seasons(cls, db_session):
        return db_session.query(cls).filter_by(season_type=SeasonType.REGULAR_SEASON).all()

    @classmethod
    def is_this_the_asg_date(cls, db_session, game_date):
        season = cls.find_by_year(db_session, game_date.year)
        return game_date == season.asg_date if season else None

    @classmethod
    def get_regular_season_map(cls, db_session):
        regular_seasons = cls.get_all_regular_seasons(db_session)
        return {s.year: s.id for s in regular_seasons}

    @classmethod
    def get_total_games_for_all_seasons(cls, db_session):
        return {s.year: s.total_games for s in cls.get_all_regular_seasons(db_session)}



from copy import deepcopy
from functools import cached_property
from pathlib import Path

from events import Events
from halo import Halo

from vigorish.cli.components import get_random_cli_color, get_random_dots_spinner
from vigorish.enums import SyncDirection, VigFile
from vigorish.tasks.base import Task
from vigorish.util.regex import URL_ID_REGEX
from vigorish.util.result import Result
from vigorish.util.sys_helpers import get_last_mod_time_utc

TIME_DIFFERENCE_SECONDS = 60


class SyncScrapedDataTask(Task):
    def __init__(self, app):
        super().__init__(app)
        self.sync_direction = None
        self.file_type = None
        self.data_set = None
        self.year = None
        self.spinner = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
        self.events = Events(
            (
                "error_occurred",
                "get_s3_objects_start",
                "get_s3_objects_complete",
                "find_out_of_sync_files_start",
                "find_out_of_sync_files_complete",
                "sync_files_start",
                "sync_files_progress",
                "sync_files_complete",
            )
        )

    @property
    def file_helper(self):
        return self.scraped_data.file_helper

    @cached_property
    def all_s3_objects(self):  # pragma: no cover
        return self.file_helper.get_all_object_keys_in_s3_bucket()

    def execute(self, sync_direction, file_type, data_set, year):
        self.get_all_s3_objects()
        result = self.find_out_of_sync_files(sync_direction, file_type, data_set, year)
        if result.failure:
            self.events.error_occurred("Error occurred analyzing which files need to be synced.")
            return result
        (out_of_sync, missing_files, outdated_files) = result.value
        if not out_of_sync:
            return Result.Ok()
        self.sync_files(sync_direction, missing_files, outdated_files, file_type, data_set, year)
        return Result.Ok()

    def get_all_s3_objects(self):
        self.events.get_s3_objects_start()
        self.all_s3_objects
        self.events.get_s3_objects_complete()

    def find_out_of_sync_files(self, sync_direction, file_type, data_set, year):
        self.events.find_out_of_sync_files_start()
        (s3_objects, local_files) = self.get_all_files_in_src_and_dest(file_type, data_set, year)
        if (sync_direction == SyncDirection.UP_TO_S3 and not local_files) or (
            sync_direction == SyncDirection.DOWN_TO_LOCAL and not s3_objects
        ):
            sync_results = (False, [], [])
            self.events.find_out_of_sync_files_complete(sync_results)
            return Result.Ok(sync_results)
        src_files = local_files if sync_direction == SyncDirection.UP_TO_S3 else s3_objects
        dest_files = s3_objects if sync_direction == SyncDirection.UP_TO_S3 else local_files
        sync_results = get_files_to_sync(src_files, dest_files)
        self.events.find_out_of_sync_files_complete(sync_results)
        return Result.Ok(sync_results)

    def get_all_files_in_src_and_dest(self, file_type, data_set, year):
        local_files = self.get_local_files(file_type, data_set, year)
        s3_objects = self.get_s3_objects(file_type, data_set, year)
        return (s3_objects, local_files)

    def get_local_files(self, file_type, data_set, year):
        folderpath = self.scraped_data.get_local_folderpath(file_type, data_set, year)
        id_regex = URL_ID_REGEX[file_type][data_set]
        local_files = filter(lambda x: id_regex.search(x.stem), Path(folderpath).glob("*.*"))
        return sorted(map(get_local_file_data, local_files), key=lambda x: x["name"])

    def get_s3_objects(self, file_type, data_set, year):
        folderpath = self.scraped_data.get_s3_folderpath(file_type, data_set, year)
        id_regex = URL_ID_REGEX[file_type][data_set]
        file_suffix = ".html" if file_type == VigFile.SCRAPED_HTML else ".json"
        s3_objects = filter(
            lambda x: folderpath in x.key and id_regex.search(Path(x.key).stem) and Path(x.key).suffix == file_suffix,
            self.all_s3_objects,
        )
        return sorted(map(get_s3_object_data, s3_objects), key=lambda x: x["name"])

    def sync_files(self, sync_direction, missing_files, outdated_files, file_type, data_set, year):
        self.file_helper.create_all_folderpaths(year)
        self.bucket_name = self.config.get_current_setting("S3_BUCKET", data_set)
        sync_files = get_sync_files(missing_files, outdated_files)
        self.events.sync_files_start(sync_files[0]["name"], 0, len(sync_files))
        for num, file in enumerate(sync_files, start=1):
            self.events.sync_files_progress(file["name"], num - 1, len(sync_files))
            (filepath, s3_key) = self.get_local_path_and_s3_key(file, file_type, data_set, year)
            self.send_file(sync_direction, filepath, s3_key)
            self.events.sync_files_progress(file["name"], num, len(sync_files))
        self.events.sync_files_complete()

    def get_local_path_and_s3_key(self, file, file_type, data_set, year):
        local_folder = self.scraped_data.get_local_folderpath(file_type, data_set, year)
        s3_folder = self.scraped_data.get_s3_folderpath(file_type, data_set, year)
        local_path = str(Path(local_folder).joinpath(file["name"]))
        s3_key = f'{s3_folder}/{file["name"]}'
        return (local_path, s3_key)

    def send_file(self, sync_direction, local_path, s3_key):  # pragma: no cover
        if sync_direction == SyncDirection.UP_TO_S3:
            self.file_helper.get_s3_bucket().upload_file(local_path, s3_key)
        if sync_direction == SyncDirection.DOWN_TO_LOCAL:
            self.file_helper.get_s3_bucket().download_file(s3_key, local_path)


def get_local_file_data(file):
    return {
        "name": file.name,
        "path": file,
        "size": file.stat().st_size,
        "last_modified": get_last_mod_time_utc(file),
    }


def get_s3_object_data(obj):
    return {
        "name": Path(obj.key).name,
        "path": Path(obj.key),
        "size": obj.size,
        "last_modified": obj.last_modified,
    }


def get_files_to_sync(src_files, dest_files):
    missing_files = find_missing_files(src_files, dest_files)
    outdated_files = find_outdated_files(src_files, dest_files)
    if missing_files or outdated_files:
        return (True, missing_files, outdated_files)
    return (False, [], [])


def find_missing_files(src_files, dest_files):
    dest_names = [f["name"] for f in dest_files if f["size"] > 0]
    return list(map(add_new_file, filter(lambda x: x["name"] not in dest_names, src_files)))


def add_new_file(file):
    file["operation"] = "ADD NEW FILE"
    return file


def find_outdated_files(src_files, dest_files):
    common_files = find_common_files(src_files, dest_files)
    return list(map(update_existing_file, filter(src_file_is_newer, common_files)))


def find_common_files(src_files, dest_files):
    src_names = {f["name"] for f in src_files if f["size"] > 0}
    dest_names = [f["name"] for f in dest_files if f["size"] > 0]
    common_names = src_names.intersection(dest_names)
    src_common_files = sorted(filter(lambda x: x["name"] in common_names, src_files), key=lambda x: x["name"])
    dest_common_files = sorted(filter(lambda x: x["name"] in common_names, dest_files), key=lambda x: x["name"])
    return zip(src_common_files, dest_common_files)


def src_file_is_newer(file_pair):
    (src, dest) = file_pair
    diff = src["last_modified"] - dest["last_modified"]
    return diff.total_seconds() > TIME_DIFFERENCE_SECONDS and src["size"] != dest["size"]


def update_existing_file(file_pair):
    (src, _) = file_pair
    src["operation"] = "UPDATE EXISTING FILE"
    return src


def get_sync_files(missing_files, outdated_files):
    sync_files = []
    if missing_files:
        sync_files = deepcopy(missing_files)
    if outdated_files:
        sync_files.extend(deepcopy(outdated_files))
    return sync_files



from functools import cached_property

import vigorish.database as db
from vigorish.util.string_helpers import fuzzy_match


class PlayerNameSearch:
    def __init__(self, db_session):
        self.db_session = db_session

    @cached_property
    def player_id_name_map(self):
        pitcher_ids = {ps.player_id_mlb for ps in self.db_session.query(db.PitchStats).all()}
        batter_ids = {bs.player_id_mlb for bs in self.db_session.query(db.BatStats).all()}
        all_ids = set(list(pitcher_ids) + list(batter_ids))
        return {player_id: db.PlayerId.find_by_mlb_id(self.db_session, player_id).mlb_name for player_id in all_ids}

    def fuzzy_match(self, query, score_cutoff=80):
        return fuzzy_match(query, self.player_id_name_map, score_cutoff)



from tqdm import tqdm

import vigorish.database as db
from vigorish.enums import SeasonType
from vigorish.util.exceptions import InvalidSeasonException
from vigorish.util.result import Result


def create_relationships(db_session):
    result = link_teams_and_seasons(db_session)
    if result.failure:
        return result

    db_session.commit()
    return Result.Ok()


def link_teams_and_seasons(db_session):
    for team in tqdm(
        db_session.query(db.Team).all(),
        desc="Linking team/season tables.....",
        unit="row",
        mininterval=0.12,
        maxinterval=5,
        unit_scale=True,
        ncols=90,
    ):
        try:
            reg_season = db.Season.find_by_year(db_session, team.year)
            post_season = db.Season.find_by_year(db_session, team.year, season_type=SeasonType.POST_SEASON)
            team.regular_season_id = reg_season.id
            team.post_season_id = post_season.id
        except InvalidSeasonException:
            continue
    return Result.Ok()
